import sys
import time
import logging
from sys import stdout
import openmm as mm
import openmm.app as app
from openmm.unit import kelvin, bar, kilocalorie_per_mole, kilojoules_per_mole
from atm.ommsystem import OMMSystemAmberRBFE
from atm.atomutils import AtomUtils

logger = logging.getLogger(__name__)


class OMMSystemAmberRBFEnoATM(OMMSystemAmberRBFE):
    def create_system(self):

        self.load_amber_system()
        self.atm_utils = AtomUtils(self.system)
        self.set_ligand_atoms()
        self.set_displacement()
        self.set_vsite_restraints()
        # set orientation restraints
        self.set_orientation_restraints()
        # set reference atoms for alignment force
        self.set_alignmentForce()
        # indexes of the atoms whose position is restrained near the initial positions
        # by a flat-bottom harmonic potential.
        self.set_positional_restraints()

        # target temperature
        temp = self.keywords.get("TEMPERATURES")
        if temp is None:
            self.temperature = 300.0 * kelvin
        elif isinstance(temp, list):
            self.temperature = float(temp[0]) * kelvin
        else:
            self.temperature = float(temp) * kelvin

        self.set_torsion_metaDbias(self.temperature)

        # do not include ATM Force, instead place the nonbonded
        # forces in what it would be the the ATMForce group
        # for the integrator
        # self.set_atmforce()
        self.atmforcegroup = self.nonbondedforcegroup  # for integrator

        self.atmforcegroup = self.free_force_group()
        import re

        nbpattern = re.compile(".*Nonbonded.*")
        for i in range(self.system.getNumForces()):
            if nbpattern.match(str(type(self.system.getForce(i)))):
                nbforce = self.system.getForce(i)
                nbforce.setForceGroup(self.atmforcegroup)
                break

        # add barostat
        pressure = 1 * bar
        self.set_barostat(self.temperature, pressure, 25)

        self.set_integrator(self.temperature, self.frictionCoeff, self.MDstepsize)


def do_mintherm(keywords, logger):
    basename = keywords.get("BASENAME")
    jobname = basename

    prmtopfile = basename + ".prmtop"
    crdfile = basename + ".inpcrd"

    # OpenMM system for minimization, thermalization, NPT, NVT
    # does not include ATM Force
    syst = OMMSystemAmberRBFEnoATM(basename, keywords, prmtopfile, crdfile, logger)
    syst.create_system()

    platform_properties = {}
    platform_name = keywords.get("OPENMM_PLATFORM")
    if platform_name is None:
        platform_name = "CUDA"
    if platform_name == "CUDA" or platform_name == "OpenCL" or platform_name == "HIP":
        platform_properties["Precision"] = "mixed"
    platform = mm.Platform.getPlatformByName(platform_name)

    simulation = app.Simulation(
        syst.topology, syst.system, syst.integrator, platform, platform_properties
    )
    simulation.context.setPositions(syst.positions)
    if syst.inpcrd.boxVectors is not None:
        simulation.context.setPeriodicBoxVectors(*syst.inpcrd.boxVectors)

    print("Using platform %s" % simulation.context.getPlatform().getName())

    print(
        "Potential energy before minimization =",
        simulation.context.getState(getEnergy=True).getPotentialEnergy(),
    )
    print("Energy minimizing the system ...")
    simulation.minimizeEnergy()
    print(
        "Potential energy after minimization =",
        simulation.context.getState(getEnergy=True).getPotentialEnergy(),
    )

    # saves minimization checkpoint
    simulation.saveState(jobname + "_min.xml")
    # saves a pdb file
    positions = simulation.context.getState(getPositions=True).getPositions()
    boxsize = simulation.context.getState().getPeriodicBoxVectors()
    simulation.topology.setPeriodicBoxVectors(boxsize)
    with open(jobname + "_min.pdb", "w") as output:
        app.PDBFile.writeFile(simulation.topology, positions, output)

    print("Thermalization ...")

    totalSteps = int(keywords.get("EQUIL_MINIM_STEPS", 150000))
    steps_per_cycle = int(keywords.get("STEPS_PER_CYCLE", 5000))
    number_of_cycles = int(totalSteps / steps_per_cycle)
    simulation.reporters.append(
        app.StateDataReporter(
            stdout,
            steps_per_cycle,
            step=True,
            potentialEnergy=True,
            temperature=True,
            volume=True,
        )
    )

    # initial temperature
    initial_temp = keywords.get("INITIAL_TEMPERATURE")
    if initial_temp is None:
        initial_temperature = 50.0 * kelvin
    else:
        initial_temperature = float(initial_temp) * kelvin

    final_temperature = syst.temperature
    delta_temperature = (final_temperature - initial_temperature) / number_of_cycles

    syst.barostat.setFrequency(999999999)  # disabled

    # MD with temperature ramp
    temperature = initial_temperature
    syst.integrator.setTemperature(temperature)
    for i in range(number_of_cycles):
        simulation.step(steps_per_cycle)
        # prepare system for new temperature
        temperature = temperature + delta_temperature
        syst.integrator.setTemperature(temperature)

    # saves thermalized checkpoint
    simulation.saveState(jobname + "_therm.xml")
    # saves a pdb file
    positions = simulation.context.getState(getPositions=True).getPositions()
    boxsize = simulation.context.getState().getPeriodicBoxVectors()
    simulation.topology.setPeriodicBoxVectors(boxsize)
    with open(jobname + "_therm.pdb", "w") as output:
        app.PDBFile.writeFile(simulation.topology, positions, output)

    print("NPT equilibration ...")

    syst.barostat.setFrequency(25)

    for i in range(number_of_cycles):
        simulation.step(steps_per_cycle)

    # saves checkpoint
    simulation.saveState(jobname + "_npt.xml")
    # saves a pdb file
    positions = simulation.context.getState(getPositions=True).getPositions()
    boxsize = simulation.context.getState().getPeriodicBoxVectors()
    simulation.topology.setPeriodicBoxVectors(boxsize)
    with open(jobname + "_npt.pdb", "w") as output:
        app.PDBFile.writeFile(simulation.topology, positions, output)

    print("NVT equilibration ...")
    syst.barostat.setFrequency(999999999)  # disabled

    # MD at constant volume
    for i in range(number_of_cycles):
        simulation.step(steps_per_cycle)

    # saves checkpoint
    simulation.saveState(jobname + "_equil.xml")
    # saves a pdb file
    positions = simulation.context.getState(getPositions=True).getPositions()
    boxsize = simulation.context.getState().getPeriodicBoxVectors()
    simulation.topology.setPeriodicBoxVectors(boxsize)
    with open(jobname + "_equil.pdb", "w") as output:
        app.PDBFile.writeFile(simulation.topology, positions, output)


def do_lambda_annealing(keywords, logger):
    basename = keywords.get("BASENAME")
    jobname = basename

    prmtopfile = basename + ".prmtop"
    crdfile = basename + ".inpcrd"

    syst = OMMSystemAmberRBFE(basename, keywords, prmtopfile, crdfile, logger)
    syst.create_system()

    platform_properties = {}
    platform_name = keywords.get("OPENMM_PLATFORM")
    if platform_name is None:
        platform_name = "CUDA"
    if platform_name == "CUDA" or platform_name == "OpenCL" or platform_name == "HIP":
        platform_properties["Precision"] = "mixed"
    platform = mm.Platform.getPlatformByName(platform_name)

    simulation = app.Simulation(
        syst.topology, syst.system, syst.integrator, platform, platform_properties
    )
    simulation.context.setPositions(syst.positions)
    if syst.inpcrd.boxVectors is not None:
        simulation.context.setPeriodicBoxVectors(*syst.inpcrd.boxVectors)

    print("Using platform %s" % simulation.context.getPlatform().getName())

    syst.barostat.setFrequency(999999999)  # disabled

    # target temperature
    temp = keywords.get("TEMPERATURES")
    if temp is None:
        temperature = 300.0 * kelvin
    elif isinstance(temp, list):
        temperature = float(temp[0]) * kelvin
    else:
        temperature = float(temp) * kelvin

    lmbd = 0.0
    lambda1 = lmbd
    lambda2 = lmbd
    alpha = 0.0 / kilocalorie_per_mole
    uh = 0.0 * kilocalorie_per_mole
    w0coeff = 0.0 * kilocalorie_per_mole
    umsc = 1000.0 * kilocalorie_per_mole
    ubcore = 500.0 * kilocalorie_per_mole
    acore = 0.062500
    direction = 1

    print("LoadState ...")
    simulation.loadState(jobname + "_equil.xml")

    # override ATM parameters
    simulation.context.setParameter(syst.atmforce.Lambda1(), lambda1)
    simulation.context.setParameter(syst.atmforce.Lambda2(), lambda2)
    simulation.context.setParameter(syst.atmforce.Alpha(), alpha * kilojoules_per_mole)
    simulation.context.setParameter(syst.atmforce.Uh(), uh / kilojoules_per_mole)
    simulation.context.setParameter(syst.atmforce.W0(), w0coeff / kilojoules_per_mole)
    simulation.context.setParameter(syst.atmforce.Umax(), umsc / kilojoules_per_mole)
    simulation.context.setParameter(
        syst.atmforce.Ubcore(), ubcore / kilojoules_per_mole
    )
    simulation.context.setParameter(syst.atmforce.Acore(), acore)
    simulation.context.setParameter(syst.atmforce.Direction(), direction)

    if syst.doMetaD:
        fgroups = {0, syst.metaDforcegroup, syst.atmforcegroup}
    else:
        fgroups = {0, syst.atmforcegroup}

    state = simulation.context.getState(getEnergy=True)
    # print("Potential Energy =", state.getPotentialEnergy())

    print("Annealing to lambda = 1/2 ...")

    totalSteps = int(keywords.get("EQUIL_ANNEALING_STEPS", 250000))
    steps_per_cycle = int(keywords.get("STEPS_PER_CYCLE", 5000))
    number_of_cycles = int(totalSteps / steps_per_cycle)
    deltalambda = (0.5 - 0.0) / float(number_of_cycles)
    simulation.reporters.append(
        app.StateDataReporter(
            stdout, steps_per_cycle, step=True, potentialEnergy=True, temperature=True
        )
    )
    simulation.reporters.append(
        app.DCDReporter(jobname + "_mdlambda.dcd", steps_per_cycle)
    )

    state = simulation.context.getState(getEnergy=True, groups=fgroups)
    print("Potential Energy =", state.getPotentialEnergy())

    binding_file = jobname + "_mdlambda.out"
    f = open(binding_file, "w")

    for i in range(number_of_cycles):
        simulation.step(steps_per_cycle)
        state = simulation.context.getState(getEnergy=True)
        pot_energy = state.getPotentialEnergy()
        (u1, u0, ebias) = syst.atmforce.getPerturbationEnergy(simulation.context)
        umcore = (
            simulation.context.getParameter(syst.atmforce.Umax()) * kilojoules_per_mole
        )
        ubcore = (
            simulation.context.getParameter(syst.atmforce.Ubcore())
            * kilojoules_per_mole
        )
        acore = simulation.context.getParameter(syst.atmforce.Acore())
        direction = simulation.context.getParameter(syst.atmforce.Direction())
        if direction > 0:
            pert_energy = syst.atm_utils.softCorePertE(u1 - u0, umcore, ubcore, acore)
        else:
            pert_energy = syst.atm_utils.softCorePertE(u0 - u1, umcore, ubcore, acore)
        l1 = simulation.context.getParameter(syst.atmforce.Lambda1())
        l2 = simulation.context.getParameter(syst.atmforce.Lambda2())
        a = simulation.context.getParameter(syst.atmforce.Alpha()) / kilojoules_per_mole
        umid = simulation.context.getParameter(syst.atmforce.Uh()) * kilojoules_per_mole
        w0 = simulation.context.getParameter(syst.atmforce.W0()) * kilojoules_per_mole
        print(
            "%f %f %f %f %f %f %f %f %f"
            % (
                temperature / kelvin,
                lmbd,
                l1,
                l2,
                a * kilocalorie_per_mole,
                umid / kilocalorie_per_mole,
                w0 / kilocalorie_per_mole,
                pot_energy / kilocalorie_per_mole,
                pert_energy / kilocalorie_per_mole,
            ),
            file=f,
        )
        f.flush()
        lmbd += deltalambda
        lambda1 += deltalambda
        lambda2 += deltalambda
        simulation.context.setParameter(syst.atmforce.Lambda1(), lambda1)
        simulation.context.setParameter(syst.atmforce.Lambda2(), lambda2)

    f.close()

    print("SaveState ...")
    simulation.saveState(jobname + "_mdlambda.xml")

    # saves a pdb file
    positions = simulation.context.getState(getPositions=True).getPositions()
    boxsize = simulation.context.getState().getPeriodicBoxVectors()
    simulation.topology.setPeriodicBoxVectors(boxsize)
    with open(jobname + "_mdlambda.pdb", "w") as output:
        app.PDBFile.writeFile(simulation.topology, positions, output)


def do_equil(keywords, logger):
    basename = keywords.get("BASENAME")
    jobname = basename

    prmtopfile = basename + ".prmtop"
    crdfile = basename + ".inpcrd"

    syst = OMMSystemAmberRBFE(basename, keywords, prmtopfile, crdfile, logger)
    syst.create_system()

    platform_properties = {}
    platform_name = keywords.get("OPENMM_PLATFORM")
    if platform_name is None:
        platform_name = "CUDA"
    if platform_name == "CUDA" or platform_name == "OpenCL" or platform_name == "HIP":
        platform_properties["Precision"] = "mixed"
    platform = mm.Platform.getPlatformByName(platform_name)

    simulation = app.Simulation(
        syst.topology, syst.system, syst.integrator, platform, platform_properties
    )
    simulation.context.setPositions(syst.positions)
    if syst.inpcrd.boxVectors is not None:
        simulation.context.setPeriodicBoxVectors(*syst.inpcrd.boxVectors)

    print("Using platform %s" % simulation.context.getPlatform().getName())

    syst.barostat.setFrequency(999999999)  # disabled

    # # target temperature
    # temp = keywords.get("TEMPERATURES")
    # if temp is None:
    #     temperature = 300.0 * kelvin
    # elif isinstance(temp, list):
    #     temperature = float(temp[0]) * kelvin
    # else:
    #     temperature = float(temp) * kelvin

    lmbd = 0.5
    lambda1 = lmbd
    lambda2 = lmbd
    alpha = 0.0 / kilocalorie_per_mole
    uh = 0.0 * kilocalorie_per_mole
    w0coeff = 0.0 * kilocalorie_per_mole
    umsc = 1000.0 * kilocalorie_per_mole
    ubcore = 500.0 * kilocalorie_per_mole
    acore = 0.062500
    direction = 1

    print("LoadState ...")
    simulation.loadState(jobname + "_mdlambda.xml")

    # override ATM parameters
    simulation.context.setParameter(syst.atmforce.Lambda1(), lambda1)
    simulation.context.setParameter(syst.atmforce.Lambda2(), lambda2)
    simulation.context.setParameter(syst.atmforce.Alpha(), alpha * kilojoules_per_mole)
    simulation.context.setParameter(syst.atmforce.Uh(), uh / kilojoules_per_mole)
    simulation.context.setParameter(syst.atmforce.W0(), w0coeff / kilojoules_per_mole)
    simulation.context.setParameter(syst.atmforce.Umax(), umsc / kilojoules_per_mole)
    simulation.context.setParameter(
        syst.atmforce.Ubcore(), ubcore / kilojoules_per_mole
    )
    simulation.context.setParameter(syst.atmforce.Acore(), acore)
    simulation.context.setParameter(syst.atmforce.Direction(), direction)

    # if syst.doMetaD:
    #     fgroups = {0, syst.metaDforcegroup, syst.atmforcegroup}
    # else:
    #     fgroups = {0, syst.atmforcegroup}

    state = simulation.context.getState(getEnergy=True)
    # print("Potential Energy =", state.getPotentialEnergy())

    print("Equilibration at lambda = 1/2 ...")

    totalSteps = int(keywords.get("EQUIL_STEPS", 150000))
    steps_per_cycle = int(keywords.get("STEPS_PER_CYCLE", 5000))
    simulation.reporters.append(
        app.StateDataReporter(
            stdout, steps_per_cycle, step=True, potentialEnergy=True, temperature=True
        )
    )
    simulation.reporters.append(app.DCDReporter(jobname + "_0.dcd", steps_per_cycle))

    state = simulation.context.getState(getEnergy=True)
    print("Potential Energy =", state.getPotentialEnergy())

    simulation.step(totalSteps)

    print("SaveState ...")
    simulation.saveState(jobname + "_0.xml")

    # saves a pdb file
    positions = simulation.context.getState(getPositions=True).getPositions()
    boxsize = simulation.context.getState().getPeriodicBoxVectors()
    simulation.topology.setPeriodicBoxVectors(boxsize)
    with open(jobname + "_0.pdb", "w") as output:
        app.PDBFile.writeFile(simulation.topology, positions, output)


def massage_keywords(keywords, restrain_solutes=True):

    # use 1 fs time step
    keywords["TIME_STEP"] = 0.001

    # restrain all solutes: receptor and ligands
    nlig2 = len(keywords.get("LIGAND2_ATOMS"))
    last_lig2_atom = int(keywords.get("LIGAND2_ATOMS")[nlig2 - 1])
    keywords["POS_RESTRAINED_ATOMS"] = [i for i in range(last_lig2_atom + 1)]


def rbfe_structprep(config_file=None):
    from atm.utils import set_directory
    from pathlib import Path
    import yaml
    import json

    if config_file is None:
        config_file = sys.argv[1]

    print("")
    print("========================================")
    print("AToM RBFE Structure Preparation         ")
    print("========================================")
    print("")
    print("Started at: " + str(time.asctime()))
    print("Input file:", config_file)
    print("")
    sys.stdout.flush()

    if config_file.endswith(".yaml"):
        with open(config_file, "r") as f:
            keywords = yaml.safe_load(f)
    elif config_file.endswith(".json"):
        with open(config_file, "r") as f:
            keywords = json.load(f)
    else:
        raise ValueError("Invalid configuration file format")

    with set_directory(Path(config_file).parent):
        restrain_solutes = True
        old_keywords = keywords.copy()
        massage_keywords(keywords, restrain_solutes)

        do_mintherm(keywords, logger)
        do_lambda_annealing(keywords, logger)

        # reestablish the restrained atoms
        if restrain_solutes:
            keywords["POS_RESTRAINED_ATOMS"] = old_keywords.get("POS_RESTRAINED_ATOMS")

        do_equil(keywords, logger)


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Specify ONE input file"

    rbfe_structprep(sys.argv[1])



import setuptools_scm
import yaml

try:
    __version__ = setuptools_scm.get_version()
except Exception:
    print("Could not get version. Defaulting to version 0")
    __version__ = "0"

with open("conda_deps.txt", "r") as f:
    deps = [line.strip() for line in f if not line.strip().startswith("#")]
    deps = [dep for dep in deps if len(dep.strip()) > 0]

for i, dep in enumerate(deps):
    parts = [part.strip() for part in dep.split("#")]
    if len(parts) > 1 and parts[1].startswith("["):
        deps[i] = {"if": parts[1][1:-1], "then": parts[0]}

# Fix conda meta.yaml
with open("package/atm/recipe_template.yaml", "r") as f:
    recipe = yaml.load(f, Loader=yaml.FullLoader)

recipe["package"]["version"] = __version__
recipe["requirements"]["run"] = deps

with open("package/atm/recipe.yaml", "w") as f:
    yaml.dump(recipe, f)



import os
from openmm import OpenMMException, Platform
from openmm.app import Simulation, StateDataReporter
from openmm.unit import kelvin, kilojoules_per_mole
from atm.utils import Timer


class OMMWorkerATM:

    def __init__(self, ommsystem, config, logger):
        self.ommsystem = ommsystem
        self.config = config
        self.logger = logger

        self.topology = self.ommsystem.topology
        self.integrator = self.ommsystem.integrator

        nodefile = self.config.get("NODEFILE")
        assert nodefile, "NODEFILE needs to be specified"
        device = (
            open(nodefile, "r").readline().split(",")[1].strip().split(":")[1].strip()
        )

        if Platform.getNumPlatforms() == 1:
            if conda_prefix := os.environ.get("CONDA_PREFIX"):
                plugin_dir = os.path.join(conda_prefix, "lib", "plugins")
                Platform.loadPluginsFromDirectory(plugin_dir)

        platform_name = self.config.get("OPENMM_PLATFORM", "CUDA")

        platform = Platform.getPlatformByName(platform_name)
        properties = {}
        if platform_name in ("CUDA", "OpenCL", "HIP"):
            properties = {"DeviceIndex": device, "Precision": "mixed"}

        self.logger.info(f"Platform: {platform_name} {properties}")

        self.simulation = Simulation(
            self.topology, self.ommsystem.system, self.integrator, platform, properties
        )
        self.context = self.simulation.context
        self.context.setPositions(self.ommsystem.positions)
        self.context.setPeriodicBoxVectors(*self.ommsystem.boxvectors)

        # one preliminary energy evaluation seems to be required to init the energy routines
        self.context.getState(getEnergy=True).getPotentialEnergy()

        # load initial state/coordinates
        basename = self.config["BASENAME"]
        self.simulation.loadState(basename + "_0.xml")

        # replace parameters loaded from the initial xml file with the values in the system
        for key, value in self.ommsystem.cparams.items():
            self.context.setParameter(key, value)

        wdir = f"cntxt_{device}"
        if not os.path.isdir(wdir):
            os.mkdir(wdir)
        self.logfile = open(os.path.join(wdir, basename), "a+")
        nprnt = int(self.config.get("PRNT_FREQUENCY"))
        self.simulation.reporters.append(
            StateDataReporter(
                self.logfile, nprnt, step=True, temperature=True, speed=True
            )
        )

    def set_state(self, par):
        self.logger.debug("ommworker.set_state")

        self.integrator.setTemperature(par["temperature"])
        self.context.setParameter(
            self.ommsystem.parameter["temperature"], par["temperature"] / kelvin
        )

        atmforce = self.ommsystem.atmforce
        self.context.setParameter(atmforce.Lambda1(), par["lambda1"])
        self.context.setParameter(atmforce.Lambda2(), par["lambda2"])
        self.context.setParameter(atmforce.Alpha(), par["alpha"] * kilojoules_per_mole)
        self.context.setParameter(atmforce.Uh(), par["uh"] / kilojoules_per_mole)
        self.context.setParameter(atmforce.W0(), par["w0"] / kilojoules_per_mole)
        self.context.setParameter(atmforce.Direction(), par["atmdirection"])

    def set_posvel(self, positions, velocities):
        self.logger.debug("ommworker.set_posvel")
        self.context.setPositions(positions)
        self.context.setVelocities(velocities)

    def get_energy(self, par):
        self.logger.debug("ommworker.get_energy")
        fgroups = {0, self.ommsystem.atmforcegroup}
        state = self.context.getState(getEnergy=True, groups=fgroups)
        pot = {}
        pot["potential_energy"] = state.getPotentialEnergy()

        (u1, u0, _) = self.ommsystem.atmforce.getPerturbationEnergy(self.context)
        umcore = (
            self.context.getParameter(self.ommsystem.atmforce.Umax())
            * kilojoules_per_mole
        )
        ubcore = (
            self.context.getParameter(self.ommsystem.atmforce.Ubcore())
            * kilojoules_per_mole
        )
        acore = self.context.getParameter(self.ommsystem.atmforce.Acore())
        if par["atmdirection"] > 0:
            pot["perturbation_energy"] = self.ommsystem.atm_utils.softCorePertE(
                u1 - u0, umcore, ubcore, acore
            )
        else:
            pot["perturbation_energy"] = self.ommsystem.atm_utils.softCorePertE(
                u0 - u1, umcore, ubcore, acore
            )
        if self.ommsystem.doMetaD:
            state = self.simulation.context.getState(
                getEnergy=True, groups={self.ommsystem.metaDforcegroup}
            )
            pot["bias_energy"] = state.getPotentialEnergy()
        else:
            pot["bias_energy"] = 0.0 * kilojoules_per_mole

        return pot

    def get_posvel(self):
        self.logger.debug("ommworker.get_posvel")
        state = self.context.getState(getPositions=True, getVelocities=True)
        return state.getPositions(asNumpy=True), state.getVelocities(asNumpy=True)

    def run(self, replica):
        assert replica.worker is self

        with Timer(self.logger.debug, "set replica state"):
            _, par = replica.get_state()
            self.set_state(par)
            self.set_posvel(replica.positions, replica.velocities)

        with Timer(self.logger.debug, "run replica"):
            nsteps = int(self.config["PRODUCTION_STEPS"])
            ntry = 5
            for _ in range(ntry):
                try:
                    self.simulation.step(nsteps)
                    break
                except OpenMMException as e:
                    self.logger.warning(f"Simulation failed: {e}")
            else:
                self.logger.error(f"Simulation failed {ntry} times!")
                raise RuntimeError(f"Simulation failed {ntry} times!")

        with Timer(self.logger.debug, "get replica state"):
            pos, vel = self.get_posvel()
            _, par = replica.get_state()
            pot = self.get_energy(par)

            replica.set_posvel(pos, vel)
            replica.set_energy(pot)
            replica.set_cycle(replica.get_cycle() + 1)
            replica.set_mdsteps(replica.get_mdsteps() + nsteps)



import sys
import copy
import openmm as mm
import openmm.app as app
from openmm.unit import (
    MOLAR_GAS_CONSTANT_R,
    picosecond,
    kelvin,
    bar,
    kilocalorie_per_mole,
    angstrom,
    kilocalories_per_mole,
    kilojoules_per_mole,
    degrees,
    amu,
    nanometer,
    femtosecond,
)
from openmm.vec3 import Vec3
from atm.atomutils import AtomUtils


# OpenMM's MTSLangevinIntegrator does not have a setTemperature method
class ATMMTSLangevinIntegrator(mm.MTSLangevinIntegrator):
    def setTemperature(self, temperature):
        self.setGlobalVariableByName("kT", MOLAR_GAS_CONSTANT_R * temperature)


class OMMSystem(object):
    def __init__(self, basename, keywords, logger):
        self.system = None
        self.topology = None
        self.positions = None
        self.boxvectors = None
        self.integrator = None
        self.barostat = None
        self.keywords = keywords
        self.basename = basename
        self.logger = logger

        # parameters stored in the openmm state
        self.parameter = {}
        self.parameter["stateid"] = "REStateId"
        self.parameter["cycle"] = "RECycle"
        self.parameter["mdsteps"] = "REMDSteps"
        # more ATM property names are in atmmetaforce

        # parameters from the cntl file
        self.cparams = {}

        self.atmforcegroup = None
        self.nonbondedforcegroup = None
        self.metaDforcegroup = None

        self.frictionCoeff = float(self.keywords.get("FRICTION_COEFF")) / picosecond
        self.MDstepsize = float(self.keywords.get("TIME_STEP")) * picosecond

        self.atmforcegroup = 2
        self.nonbondedforcegroup = 1
        self.metaDforcegroup = 3

        self.doMetaD = False

    def _exit(self, message):
        """Print and flush a message to stdout and then exit."""
        self.logger.error(message)
        sys.stdout.flush()
        sys.exit(1)


class OMMSystemAmber(OMMSystem):
    def __init__(self, basename, keywords, prmtopfile, crdfile, logger):
        super().__init__(basename, keywords, logger)
        self.prmtopfile = prmtopfile
        self.crdfile = crdfile
        self.parameter["temperature"] = "RETemperature"
        self.parameter["potential_energy"] = "REPotEnergy"

    def load_amber_system(self):
        """
        sets the value of
        prmtop : Amber topology file
        inpcrd : Amber coordinates
        system : creates a OpenMM system with topology, coordinates
        topology : defines the OpenMM topology of the system
        positions : defines positions of all atoms of the system in OpenMM
        boxvectors : stores the dimension of the simulation box

        """
        self.prmtop = app.AmberPrmtopFile(self.prmtopfile)
        self.inpcrd = app.AmberInpcrdFile(self.crdfile)
        if self.keywords.get("HMASS") is not None:
            hmass = float(self.keywords.get("HMASS")) * amu
        else:
            hmass = 1.0 * amu
        self.system = self.prmtop.createSystem(
            nonbondedMethod=app.PME,
            nonbondedCutoff=0.9 * nanometer,
            constraints=app.HBonds,
            hydrogenMass=hmass,
        )
        self.topology = self.prmtop.topology
        self.positions = self.inpcrd.positions
        self.boxvectors = self.inpcrd.boxVectors

        if nnp_model := self.keywords.get("NNP_MODEL"):
            from openmmml import MLPotential

            self.logger.info("Initialize NNP/MM")

            group_indices = []
            for i in [1, 2]:
                atom_indices = list(map(int, self.keywords[f"LIGAND{i}_ATOMS"]))
                self.logger.info(f"Ligand{i}: NNP atom indices: {atom_indices}")
                self.logger.info(f"Ligand{i}: number of NNP atoms: {len(atom_indices)}")
                assert len(atom_indices) > 0
                group_indices.append(atom_indices)

            self.logger.info(f"NNP model: {nnp_model}")
            nnp = None
            if nnp_model.startswith("TorchMD-NET"):
                from atm import atom_nnp_wrapper

                nnp_file = self.keywords["NNP_FILE"]
                self.logger.info(f"NNP file: {nnp_file}")
                max_num_neighbors = self.keywords["NNP_MAX_NUM_NEIGHBORS"]
                self.logger.info(f"NNP max num neighbors: {max_num_neighbors}")
                nnp = MLPotential(
                    nnp_model,
                    model_file=nnp_file,
                    group_indices=group_indices,
                    max_num_neighbors=max_num_neighbors,
                    use_cuda_graphs=True,
                )
            else:
                nnp = MLPotential(nnp_model)

            all_atom_indices = group_indices[0] + group_indices[1]
            self.system = nnp.createMixedSystem(
                self.topology, self.system, all_atom_indices, removeConstraints=False
            )

    def set_barostat(self, temperature, pressure, frequency):
        """
        sets the system Barostat; Currently applies the MonteCarlo Barostat

        Requires: self,
        temperature
        pressure : eg. 1*bar
        frequency : 0 - disable the barostat

        """
        self.barostat = mm.MonteCarloBarostat(pressure, temperature)
        self.barostat.setFrequency(frequency)
        self.system.addForce(self.barostat)

    def free_force_group(self):
        freeGroups = set(range(32)) - set(
            force.getForceGroup() for force in self.system.getForces()
        )
        if len(freeGroups) == 0:
            self._exit(
                "Cannot find a free force group. "
                "The maximum number (32) of the force groups is already used."
            )
        return max(freeGroups)

    def set_integrator(
        self,
        temperature,
        frictionCoeff,
        MDstepsize,
        defaultMDstepsize=0.001 * picosecond,
    ):
        # place non-bonded force in group 1, assume all other bonded forces are in group 0
        nonbonded = [
            f for f in self.system.getForces() if isinstance(f, mm.NonbondedForce)
        ][0]
        self.nonbondedforcegroup = self.free_force_group()
        nonbonded.setForceGroup(self.nonbondedforcegroup)
        # set the multiplicity of the calculation of bonded forces so that they are evaluated at least once every 1 fs (default time-step)
        bonded_frequency = max(1, int(round(MDstepsize / defaultMDstepsize)))
        self.logger.info(
            "Running with a %f fs time-step with bonded forces integrated %d times per time-step"
            % (MDstepsize / femtosecond, bonded_frequency)
        )
        if self.doMetaD:
            fgroups = [
                (0, bonded_frequency),
                (self.metaDforcegroup, bonded_frequency),
                (self.nonbondedforcegroup, 1),
            ]
        else:
            fgroups = [(0, bonded_frequency), (self.nonbondedforcegroup, 1)]
        self.integrator = ATMMTSLangevinIntegrator(
            temperature, frictionCoeff, MDstepsize, fgroups
        )
        self.integrator.setConstraintTolerance(0.00001)

    def set_positional_restraints(self):
        # indexes of the atoms whose position is restrained near the initial positions
        # by a flat-bottom harmonic potential.
        posrestr_atoms_list = self.keywords.get("POS_RESTRAINED_ATOMS")
        self.posrestrForce = None
        if posrestr_atoms_list is not None:
            posrestr_atoms = [int(i) for i in posrestr_atoms_list]
            fc = float(self.keywords.get("POSRE_FORCE_CONSTANT")) * (
                kilocalorie_per_mole / angstrom**2
            )
            tol = float(self.keywords.get("POSRE_TOLERANCE")) * angstrom
            self.posrestrForce = self.atm_utils.addPosRestraints(
                posrestr_atoms, self.positions, fc, tol
            )

    def set_torsion_metaDbias(self, temperature):
        import yaml

        if self.keywords.get("METADBIAS_DIR") is None:
            return
        bias_dirs = self.keywords.get("METADBIAS_DIR")
        bias_offsets = self.keywords.get("METADBIAS_IDXOFFSET")

        for mdir, offset in zip(bias_dirs, bias_offsets):
            yamlfile = "%s/%s.yaml" % (mdir, mdir)
            with open(yamlfile, "r") as f:
                keywords = yaml.safe_load(f)

            # metadynamics settings
            bias_factor = float(
                keywords.get("METADBIAS_FACTOR")
            )  # this is (T+DeltaT)/T
            bias_height = (
                float(keywords.get("METADBIAS_GHEIGHT")) * kilocalorie_per_mole
            )  # height of each gaussian
            bias_frequency = int(
                keywords.get("METADBIAS_FREQUENCY")
            )  # steps in between gaussian depositions
            bias_savefrequency = int(
                keywords.get("METADBIAS_SAVEFREQUENCY")
            )  # steps in between checkpointing of bias potential

            # bias force settings
            torsions = keywords.get("METADBIAS_TORSIONS")
            ndim = len(torsions.keys())

            gaussian_width = keywords.get("METADBIAS_GWIDTH")
            angle_min = keywords.get("METADBIAS_MINANGLE")
            angle_max = keywords.get("METADBIAS_MAXANGLE")
            ngrid = keywords.get("METADBIAS_NGRID")
            periodic = keywords.get("METADBIAS_PERIODIC")

            torForce = []
            biasvar = []

            for t in range(ndim):
                torForce.append(mm.CustomTorsionForce("theta"))
                p = torsions[str(t)]
                gw = float(gaussian_width[t]) * kilocalorie_per_mole
                amin = float(angle_min[t]) * degrees
                amax = float(angle_max[t]) * degrees
                per = int(periodic[t]) > 0
                ng = int(ngrid[t])
                dp = int(offset)
                torForce[t].addTorsion(
                    int(p[0]) + dp, int(p[1]) + dp, int(p[2]) + dp, int(p[3]) + dp
                )
                biasvar.append(
                    app.metadynamics.BiasVariable(torForce[t], amin, amax, gw, per, ng)
                )

            metaD = app.metadynamics.Metadynamics(
                self.system,
                biasvar,
                temperature,
                bias_factor,
                bias_height,
                bias_frequency,
                bias_savefrequency,
                mdir,
            )
            self.metaDforcegroup = metaD._force.getForceGroup()
        self.doMetaD = True


class OMMSystemAmberRBFE(OMMSystemAmber):
    def __init__(self, basename, keywords, prmtopfile, crdfile, logger):
        super().__init__(basename, keywords, prmtopfile, crdfile, logger)

        self.parameter["perturbation_energy"] = "REPertEnergy"
        self.parameter["atmintermediate"] = "REAlchemicalIntermediate"
        self.parameter["bias_energy"] = "BiasEnergy"
        self.atmforce = None
        self.lig1_atoms = None
        self.lig2_atoms = None
        self.displ = None

    def set_ligand_atoms(self):
        lig1_atoms_in = self.keywords.get("LIGAND1_ATOMS")  # indexes of ligand1 atoms
        lig2_atoms_in = self.keywords.get("LIGAND2_ATOMS")  # indexes of ligand2 atoms
        if lig1_atoms_in is not None:
            self.lig1_atoms = lig1_atoms_in
        else:
            msg = "Error: LIGAND1_ATOMS is required"
            self._exit(msg)
        if lig2_atoms_in is not None:
            self.lig2_atoms = lig2_atoms_in
        else:
            msg = "Error: LIGAND2_ATOMS is required"
            self._exit(msg)

    def set_displacement(self):
        # set displacements and offsets for ligand 1 and ligand 2
        if self.keywords.get("DISPLACEMENT") is not None:
            self.displ = [
                float(displ) for displ in self.keywords.get("DISPLACEMENT")
            ] * angstrom
            self.lig1offset = [
                float(0.0 * offset) for offset in self.displ / angstrom
            ] * angstrom
            self.lig2offset = [
                float(offset) for offset in self.displ / angstrom
            ] * angstrom
        else:
            msg = "DISPLACEMENT is required"
            self._exit(msg)

    def set_vsite_restraints(self):
        # ligand 1 Vsite restraint
        lig1_atom_restr = self.keywords.get(
            "LIGAND1_CM_ATOMS"
        )  # indexes of ligand atoms for CM-CM Vsite restraint

        # ligand 2 Vsite restraint
        lig2_atom_restr = self.keywords.get(
            "LIGAND2_CM_ATOMS"
        )  # indexes of ligand atoms for CM-CM Vsite restraint

        # Vsite restraint receptor atoms
        cm_rcpt_atoms = self.keywords.get(
            "RCPT_CM_ATOMS"
        )  # indexes of rcpt atoms for CM-CM Vsite restraint
        if cm_rcpt_atoms is None:
            cm_rcpt_atoms = self.keywords.get("REST_LIGAND_CMREC_ATOMS")
        if cm_rcpt_atoms is not None:
            rcpt_atom_restr = [int(i) for i in cm_rcpt_atoms]
        else:
            rcpt_atom_restr = None

        cmrestraints_present = (
            (rcpt_atom_restr is not None)
            and (lig1_atom_restr is not None)
            and (lig2_atom_restr is not None)
        )

        self.vsiterestraintForce1 = None
        self.vsiterestraintForce2 = None
        if cmrestraints_present:
            cmkf = float(self.keywords.get("CM_KF"))
            kf = (
                cmkf * kilocalorie_per_mole / angstrom**2
            )  # force constant for Vsite CM-CM restraint
            cmtol = float(self.keywords.get("CM_TOL"))
            r0 = cmtol * angstrom  # radius of Vsite sphere

            # Vsite restraints for ligands 1 and 2
            self.vsiterestraintForce1 = self.atm_utils.addVsiteRestraintForceCMCM(
                lig_cm_particles=lig1_atom_restr,
                rcpt_cm_particles=rcpt_atom_restr,
                kfcm=kf,
                tolcm=r0,
                offset=self.lig1offset,
            )
            self.vsiterestraintForce2 = self.atm_utils.addVsiteRestraintForceCMCM(
                lig_cm_particles=lig2_atom_restr,
                rcpt_cm_particles=rcpt_atom_restr,
                kfcm=kf,
                tolcm=r0,
                offset=self.lig2offset,
            )

    def set_orientation_restraints(self):
        # orientational VSite restraints
        # the indexes of the groups of atoms that define the internal reference frame of the ligand
        lig1_frame_groups = None
        lig1_frame_groups_inp = self.keywords.get("LIGAND1_VSITE_FRAMEGROUPS")
        if lig1_frame_groups_inp is not None:
            lig1_frame_groups = []
            for i in range(3):
                lig1_frame_groups.append(
                    [int(j) for j in lig1_frame_groups_inp[str(i)]]
                )
        lig2_frame_groups = None
        lig2_frame_groups_inp = self.keywords.get("LIGAND2_VSITE_FRAMEGROUPS")
        if lig2_frame_groups_inp is not None:
            lig2_frame_groups = []
            for i in range(3):
                lig2_frame_groups.append(
                    [int(j) for j in lig2_frame_groups_inp[str(i)]]
                )
        # the indexes of the groups of atoms that define the internal reference frame of the receptor
        rcpt_frame_groups = None
        rcpt_frame_groups_inp = self.keywords.get("RCPT_VSITE_FRAMEGROUPS")
        if rcpt_frame_groups_inp is not None:
            rcpt_frame_groups = []
            for i in range(3):
                rcpt_frame_groups.append(
                    [int(j) for j in rcpt_frame_groups_inp[str(i)]]
                )
        if rcpt_frame_groups is not None:
            kftheta = self.keywords.get("VSITE_KFTHETA_LIG1")
            theta0 = self.keywords.get("VSITE_THETA0_LIG1")
            thetatol = self.keywords.get("VSITE_THETATOL_LIG1")
            kfphi = self.keywords.get("VSITE_KFPHI_LIG1")
            phi0 = self.keywords.get("VSITE_PHI0_LIG1")
            phitol = self.keywords.get("VSITE_PHITOL_LIG1")
            kfpsi = self.keywords.get("VSITE_KFPSI_LIG1")
            psi0 = self.keywords.get("VSITE_PSI0_LIG1")
            psitol = self.keywords.get("VSITE_PSITOL_LIG1")
            kftheta = (
                kftheta if kftheta is None else float(kftheta) * kilocalories_per_mole
            )
            theta0 = theta0 if theta0 is None else float(theta0) * degrees
            thetatol = thetatol if thetatol is None else float(thetatol) * degrees
            kfphi = (
                kfphi
                if kfphi is None
                else float(kfphi) * (kilocalories_per_mole / degrees**2)
            )
            phi0 = phi0 if phi0 is None else float(phi0) * degrees
            phitol = phitol if phitol is None else float(phitol) * degrees
            kfpsi = (
                kfpsi
                if kfpsi is None
                else float(kfpsi) * (kilocalories_per_mole / degrees**2)
            )
            psi0 = psi0 if psi0 is None else float(psi0) * degrees
            psitol = psitol if psitol is None else float(psitol) * degrees
            if lig1_frame_groups is not None:
                self.atm_utils.addVsiteRestraintForceCMAngles(
                    lig1_frame_groups,
                    rcpt_frame_groups,
                    kftheta,
                    theta0,
                    thetatol,
                    kfphi,
                    phi0,
                    phitol,
                    kfpsi,
                    psi0,
                    psitol,
                )
            kftheta = self.keywords.get("VSITE_KFTHETA_LIG2")
            theta0 = self.keywords.get("VSITE_THETA0_LIG2")
            thetatol = self.keywords.get("VSITE_THETATOL_LIG2")
            kfphi = self.keywords.get("VSITE_KFPHI_LIG2")
            phi0 = self.keywords.get("VSITE_PHI0_LIG2")
            phitol = self.keywords.get("VSITE_PHITOL_LIG2")
            kfpsi = self.keywords.get("VSITE_KFPSI_LIG2")
            psi0 = self.keywords.get("VSITE_PSI0_LIG2")
            psitol = self.keywords.get("VSITE_PSITOL_LIG2")
            kftheta = (
                kftheta if kftheta is None else float(kftheta) * kilocalories_per_mole
            )
            theta0 = theta0 if theta0 is None else float(theta0) * degrees
            thetatol = thetatol if thetatol is None else float(thetatol) * degrees
            kfphi = (
                kfphi
                if kfphi is None
                else float(kfphi) * (kilocalories_per_mole / degrees**2)
            )
            phi0 = phi0 if phi0 is None else float(phi0) * degrees
            phitol = phitol if phitol is None else float(phitol) * degrees
            kfpsi = (
                kfpsi
                if kfpsi is None
                else float(kfpsi) * (kilocalories_per_mole / degrees**2)
            )
            psi0 = psi0 if psi0 is None else float(psi0) * degrees
            psitol = psitol if psitol is None else float(psitol) * degrees
            if lig2_frame_groups is not None:
                self.atm_utils.addVsiteRestraintForceCMAngles(
                    lig2_frame_groups,
                    rcpt_frame_groups,
                    kftheta,
                    theta0,
                    thetatol,
                    kfphi,
                    phi0,
                    phitol,
                    kfpsi,
                    psi0,
                    psitol,
                )

    def set_alignmentForce(self):
        """
        set reference atoms for adding the alignment force

        """
        refatoms1_cntl = self.keywords.get("ALIGN_LIGAND1_REF_ATOMS")
        refatoms2_cntl = self.keywords.get("ALIGN_LIGAND2_REF_ATOMS")

        if refatoms1_cntl is None or refatoms2_cntl is None:
            return

        self.refatoms1 = [int(refatoms1) for refatoms1 in refatoms1_cntl]
        lig1_ref_atoms = [self.refatoms1[i] + self.lig1_atoms[0] for i in range(3)]
        self.refatoms2 = [int(refatoms2) for refatoms2 in refatoms2_cntl]
        lig2_ref_atoms = [self.refatoms2[i] + self.lig2_atoms[0] for i in range(3)]

        # add alignment force
        self.atm_utils.addAlignmentForce(
            liga_ref_particles=lig1_ref_atoms,
            ligb_ref_particles=lig2_ref_atoms,
            kfdispl=float(self.keywords.get("ALIGN_KF_SEP"))
            * kilocalorie_per_mole
            / angstrom**2,
            ktheta=float(self.keywords.get("ALIGN_K_THETA")) * kilocalorie_per_mole,
            kpsi=float(self.keywords.get("ALIGN_K_PSI")) * kilocalorie_per_mole,
            offset=self.lig2offset,
        )

    def set_integrator(
        self,
        temperature,
        frictionCoeff,
        MDstepsize,
        defaultMDstepsize=0.001 * picosecond,
    ):

        integrator = self.keywords.get("INTEGRATOR")
        if integrator in ("MTS", None):
            self.logger.info("Integrator: MTS")
            # set the multiplicity of the calculation of bonded forces so that they are evaluated at least once every 1 fs (default time-step)
            bonded_frequency = max(1, int(round(MDstepsize / defaultMDstepsize)))
            self.logger.info(
                "Running with a %f fs time-step with bonded forces integrated %d times per time-step"
                % (MDstepsize / femtosecond, bonded_frequency)
            )
            if self.doMetaD:
                fgroups = [
                    (0, bonded_frequency),
                    (self.metaDforcegroup, bonded_frequency),
                    (self.atmforcegroup, 1),
                ]
            else:
                fgroups = [(0, bonded_frequency), (self.atmforcegroup, 1)]
            self.integrator = ATMMTSLangevinIntegrator(
                temperature, frictionCoeff, MDstepsize, fgroups
            )

        elif integrator == "Langevin":
            self.logger.info(f"Integrator: {integrator}")
            self.integrator = mm.LangevinIntegrator(
                temperature, frictionCoeff, MDstepsize
            )
            self.integrator.setIntegrationForceGroups(
                {0, self.metaDforcegroup, self.atmforcegroup}
            )

            self.logger.info(f"Temperature: {self.integrator.getTemperature()}")
            self.logger.info(f"Friction: {self.integrator.getFriction()}")

        elif integrator == "LangevinMiddle":
            self.logger.info(f"Integrator: {integrator}")
            self.integrator = mm.LangevinMiddleIntegrator(
                temperature, frictionCoeff, MDstepsize
            )
            self.integrator.setIntegrationForceGroups(
                {0, self.metaDforcegroup, self.atmforcegroup}
            )

            self.logger.info(f"Temperature: {self.integrator.getTemperature()}")
            self.logger.info(f"Friction: {self.integrator.getFriction()}")

        else:
            raise ValueError(f"Unknown integrator: {integrator}")

        self.logger.info(f"Step size: {self.integrator.getStepSize()}")
        self.logger.info(
            f"Constraint tolerance: {self.integrator.getConstraintTolerance()}"
        )

    def set_atmforce(self):
        # these define the state and will be overriden in set_state()
        lmbd = 0.0
        lambda1 = lmbd
        lambda2 = lmbd
        alpha = 0.0 / kilocalorie_per_mole
        uh = 0.0 * kilocalorie_per_mole
        w0coeff = 0.0 * kilocalorie_per_mole
        direction = 1.0

        # soft-core parameters are fixed (the same in all states)
        umsc = float(self.keywords.get("UMAX")) * kilocalorie_per_mole
        ubcore = self.keywords.get("UBCORE")
        if ubcore:
            ubcore = float(ubcore) * kilocalorie_per_mole
        else:
            ubcore = 0.0 * kilocalorie_per_mole
        acore = float(self.keywords.get("ACORE"))

        self.atmforce = mm.ATMForce(
            lambda1,
            lambda2,
            alpha * kilojoules_per_mole,
            uh / kilojoules_per_mole,
            w0coeff / kilojoules_per_mole,
            umsc / kilojoules_per_mole,
            ubcore / kilojoules_per_mole,
            acore,
            direction,
        )

        # adds nonbonded Force from the system to the ATMForce
        import re

        nbpattern = re.compile(".*Nonbonded.*")
        for i in range(self.system.getNumForces()):
            if nbpattern.match(str(type(self.system.getForce(i)))):
                self.atmforce.addForce(copy.copy(self.system.getForce(i)))

                # https://github.com/openmm/openmm/issues/4395
                # rather then removing the nonbonded force, disable it by assigning a force
                # group not included in the MTS integrator. This way it can do atom reordering.
                # self.system.removeForce(i)
                self.nonbondedforcegroup = self.free_force_group()
                self.system.getForce(i).setForceGroup(self.nonbondedforcegroup)
                break

        # adds atoms to ATMForce
        for i in range(self.topology.getNumAtoms()):
            self.atmforce.addParticle(Vec3(0.0, 0.0, 0.0))
        for i in self.lig1_atoms:
            self.atmforce.setParticleParameters(
                i, Vec3(self.displ[0], self.displ[1], self.displ[2]) / nanometer
            )
        for i in self.lig2_atoms:
            self.atmforce.setParticleParameters(
                i, -Vec3(self.displ[0], self.displ[1], self.displ[2]) / nanometer
            )

        # assign a group to ATMForce for multiple time-steps
        self.atmforcegroup = self.free_force_group()
        self.atmforce.setForceGroup(self.atmforcegroup)

        # add ATMForce to the system
        self.system.addForce(self.atmforce)

        # these are the global parameters specified in the cntl files that need to be reset
        # by the worker after reading the first configuration
        self.cparams[self.atmforce.Umax()] = umsc / kilojoules_per_mole
        self.cparams[self.atmforce.Ubcore()] = ubcore / kilojoules_per_mole
        self.cparams[self.atmforce.Acore()] = acore

    def create_system(self):

        self.load_amber_system()
        self.atm_utils = AtomUtils(self.system)
        self.set_ligand_atoms()
        self.set_displacement()
        self.set_vsite_restraints()
        # set orientation restraints
        self.set_orientation_restraints()
        # set reference atoms for alignment force
        self.set_alignmentForce()
        # indexes of the atoms whose position is restrained near the initial positions
        # by a flat-bottom harmonic potential.
        self.set_positional_restraints()
        # temperature is part of the state and is maybe overriden in set_state()
        temperature = 300 * kelvin

        self.set_torsion_metaDbias(temperature)

        self.set_atmforce()

        # add barostat
        pressure = 1 * bar
        self.set_barostat(temperature, pressure, 0)
        # hack to store ASyncRE quantities in the openmm State
        sforce = mm.CustomBondForce("1")
        for name in self.parameter:
            sforce.addGlobalParameter(self.parameter[name], 0)
        self.system.addForce(sforce)

        self.set_integrator(temperature, self.frictionCoeff, self.MDstepsize)



import string
import random
import numpy as np
from openmm import version as ommversion
import openmm as mm
from openmm.unit import (
    kilocalorie_per_mole,
    kilocalories_per_mole,
    angstrom,
    kilojoule_per_mole,
    nanometer,
    radians,
)
import math


class AtomUtils(object):
    """
    ATM Force python utilities module
    """

    def __init__(self, system, fix_zero_LJparams=True):
        """
        Creates an ATMMetaForceUtils object.

        Parameters
        ----------
        system : System
            The OpenMM System to operate upon
        fix_zero_LJparams: bool
            Whether to fix atoms with zero LJ parameters

        Returns
        -------
        An ATMMetaForceUtils object

        """

        self.system = system

        self.CMCMDistForce = None
        self.CMAngleThetaForce = None
        self.CMAnglePhiForce = None
        self.CMAnglePsiForce = None

        self.major_ommversion = int(ommversion.version.split(".")[0])
        self.minor_ommversion = int(ommversion.version.split(".")[1])

        if fix_zero_LJparams:
            import re

            nbpattern = re.compile(".*Nonbonded.*")
            for force in self.system.getForces():
                if nbpattern.match(str(type(force))):
                    self.fixZeroLJParams(force)

    #
    # Adds restrains to keep the the CM of the ligand atoms around the CM of the receptor atoms
    # plus an offset, within a tolerance, with a flat-bottom harmonic potential
    def addVsiteRestraintForceCMCM(
        self,
        lig_cm_particles=None,
        rcpt_cm_particles=None,
        kfcm=0.0 * kilocalorie_per_mole / angstrom**2,
        tolcm=0.0 * angstrom,
        offset=[0.0, 0.0, 0.0] * angstrom,
    ):
        """
        Adds a flat-bottom quadratic potential restraint to keep the the CM of the ligand atoms
        near the CM of the receptor atoms. This potential is used to define the spatial extent of
        the binding site.

        Parameters
        ----------
        lig_cm_particles: list of ints
            List of the indexes of the atoms that define the CM of the ligand molecule
        rcpt_cm_particles : list of ints
            List of the indexes of the atoms that define the CM of the receptor molecule
        kfcm : Quantity in energy/distance^2 units
            Force constant of the quadratic potential
        tolcm : Quantity in distance units
            Tolerance of the flat-bottom restraint
        offset : List of 3 Quantity's in distance units
            Center of the restraint. The potential is applied to the CM-CM distance after subtracting offset.
            It is used to keep the CM-CM distance near offset.

        Returns
        -------
        The CM-CM restraint Force object added to the OpenMM system

        """
        assert len(lig_cm_particles) > 0
        assert len(rcpt_cm_particles) > 0

        expr = "(kfcm/2)*step(d12-tolcm)*(d12-tolcm)^2 "
        expr += " ; d12 = sqrt((x1 - offx - x2)^2 + (y1 - offy - y2)^2 + (z1 - offz - z2)^2 ) ; "
        force = None
        if self.CMCMDistForce is None:
            self.CMCMDistForce = mm.CustomCentroidBondForce(2, expr)
            force = self.CMCMDistForce
            force.addPerBondParameter("kfcm")
            force.addPerBondParameter("tolcm")
            force.addPerBondParameter("offx")
            force.addPerBondParameter("offy")
            force.addPerBondParameter("offz")
            numgroups = 0
            self.system.addForce(force)
        else:
            force = self.CMCMDistForce
            numgroups = force.getNumGroups()
        force.addGroup(lig_cm_particles)  # g1 CM of lig
        force.addGroup(rcpt_cm_particles)  # g2 CM of rcpt
        kfc = kfcm / (kilojoule_per_mole / nanometer**2)
        tolc = tolcm / nanometer
        offv = offset / nanometer
        offx = offv[0]
        offy = offv[1]
        offz = offv[2]
        groups = [numgroups + 0, numgroups + 1]
        params = [kfc, tolc, offx, offy, offz]
        force.addBond(groups, np.array(params, dtype=np.double))
        return force

    def _roundExpression(self, n, x):
        return (
            "{n} = select(step({x} - floor({x})  - 0.5), ceil({x}), floor({x}))".format(
                n=n, x=x
            )
        )

    def _wrapExpression(self, w, x, r):
        letters = string.ascii_lowercase
        n = "n" + "".join(random.choice(letters) for i in range(4))
        expr = "{w} = {x} - {r}*{n} ; ".format(w=w, x=x, r=r, n=n) + " ; "
        expr += self._roundExpression("{n}".format(n=n), "({x}/{r})".format(x=x, r=r))
        return expr

    def _dotExpression(self, dot, x1, y1, z1, x2, y2, z2):
        return "{dot} = {x1}*{x2} + {y1}*{y2} + {z1}*{z2}".format(
            dot=dot, x1=x1, y1=y1, z1=z1, x2=x2, y2=y2, z2=z2
        )

    def _unitvExpression(self, xu, yu, zu, x, y, z):
        letters = string.ascii_lowercase
        d = "d" + "".join(random.choice(letters) for i in range(4))
        return "{xu} = {x}/{d} ; {yu} = {y}/{d} ; {zu} = {z}/{d} ; {d} = sqrt({x}*{x}+{y}*{y}+{z}*{z})".format(
            d=d, xu=xu, yu=yu, zu=zu, x=x, y=y, z=z
        )

    def _diffvExpression(self, dx, dy, dz, x2, y2, z2, x1, y1, z1):
        return "{dx} = {x2} - {x1} ; {dy} = {y2} - {y1} ; {dz} = {z2} - {z1} ".format(
            dx=dx, dy=dy, dz=dz, x2=x2, y2=y2, z2=z2, x1=x1, y1=y1, z1=z1
        )

    def _crossExpression(self, xc, yc, zc, x1, y1, z1, x2, y2, z2):
        return (
            "{xc} =  {y1}*{z2} - {y2}*{z1} ; "
            "{yc} = -{x1}*{z2} + {x2}*{z1} ; "
            "{zc} =  {x1}*{y2} - {x2}*{y1}   "
        ).format(xc=xc, yc=yc, zc=zc, x1=x1, y1=y1, z1=z1, x2=x2, y2=y2, z2=z2)

    def _dihedralExpression(self, phi, x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4):
        if self.major_ommversion >= 7 and self.minor_ommversion >= 6:
            return (
                phi
                + "= pointdihedral("
                + x1
                + ","
                + y1
                + ","
                + z1
                + ","
                + x2
                + ","
                + y2
                + ","
                + z2
                + ","
                + x3
                + ","
                + y3
                + ","
                + z3
                + ","
                + x4
                + ","
                + y4
                + ","
                + z4
                + ")"
            )
        else:
            letters = string.ascii_lowercase
            v1 = "v1" + "".join(random.choice(letters) for i in range(4))
            v2 = "v2" + "".join(random.choice(letters) for i in range(4))
            v2u = "v2u" + "".join(random.choice(letters) for i in range(4))
            v3 = "v3" + "".join(random.choice(letters) for i in range(4))
            n1 = "n1" + "".join(random.choice(letters) for i in range(4))
            n1u = "n1u" + "".join(random.choice(letters) for i in range(4))
            n2 = "n2" + "".join(random.choice(letters) for i in range(4))
            n2u = "n2u" + "".join(random.choice(letters) for i in range(4))
            m1 = "m1" + "".join(random.choice(letters) for i in range(4))
            m1u = "m1u" + "".join(random.choice(letters) for i in range(4))
            cx = "cx" + "".join(random.choice(letters) for i in range(4))
            cy = "cy" + "".join(random.choice(letters) for i in range(4))
            expr = "{phi} = atan2({cy} , {cx}) ; ".format(phi=phi, cx=cx, cy=cy)
            expr += (
                self._dotExpression(
                    cx, "x" + n1u, "y" + n1u, "z" + n1u, "x" + n2u, "y" + n2u, "z" + n2u
                )
                + " ; "
            )
            expr += (
                self._dotExpression(
                    cy, "x" + m1u, "y" + m1u, "z" + m1u, "x" + n2u, "y" + n2u, "z" + n2u
                )
                + " ; "
            )
            expr += (
                self._unitvExpression(
                    "x" + m1u, "y" + m1u, "z" + m1u, "x" + m1, "y" + m1, "z" + m1
                )
                + " ; "
            )
            expr += (
                self._crossExpression(
                    "x" + m1,
                    "y" + m1,
                    "z" + m1,
                    "x" + v2u,
                    "y" + v2u,
                    "z" + v2u,
                    "x" + n1u,
                    "y" + n1u,
                    "z" + n1u,
                )
                + " ; "
            )
            expr += (
                self._unitvExpression(
                    "x" + n1u, "y" + n1u, "z" + n1u, "x" + n1, "y" + n1, "z" + n1
                )
                + " ; "
            )
            expr += (
                self._unitvExpression(
                    "x" + n2u, "y" + n2u, "z" + n2u, "x" + n2, "y" + n2, "z" + n2
                )
                + " ; "
            )
            expr += (
                self._unitvExpression(
                    "x" + v2u, "y" + v2u, "z" + v2u, "x" + v2, "y" + v2, "z" + v2
                )
                + " ; "
            )
            expr += (
                self._crossExpression(
                    "x" + n2,
                    "y" + n2,
                    "z" + n2,
                    "x" + v2,
                    "y" + v2,
                    "z" + v2,
                    "x" + v3,
                    "y" + v3,
                    "z" + v3,
                )
                + " ; "
            )
            expr += (
                self._crossExpression(
                    "x" + n1,
                    "y" + n1,
                    "z" + n1,
                    "x" + v1,
                    "y" + v1,
                    "z" + v1,
                    "x" + v2,
                    "y" + v2,
                    "z" + v2,
                )
                + " ; "
            )
            expr += (
                self._diffvExpression(
                    "x" + v1, "y" + v1, "z" + v1, x2, y2, z2, x1, y1, z1
                )
                + " ; "
            )
            expr += (
                self._diffvExpression(
                    "x" + v2, "y" + v2, "z" + v2, x3, y3, z3, x2, y2, z2
                )
                + " ; "
            )
            expr += self._diffvExpression(
                "x" + v3, "y" + v3, "z" + v3, x4, y4, z4, x3, y3, z3
            )
            return expr

    def addAlignmentForce(
        self,
        liga_ref_particles=None,
        ligb_ref_particles=None,
        kfdispl=0.0 * kilocalorie_per_mole / angstrom**2,
        ktheta=0.0 * kilocalorie_per_mole,
        kpsi=0.0 * kilocalorie_per_mole,
        offset=[0.0, 0.0, 0.0] * angstrom,
    ):
        """
        Adds Forces to keep two ligands aligned based on reference frames defined by two sets of 3 reference atoms
        (a1, a2, a3) and (b1, b2, b3) for ligands a and b, respectively.

        See https://pubs.acs.org/doi/10.1021/acs.jcim.1c01129

        Adds a quadratic potential on the a1-b1 distance, minus an offset

        Adds a potential of the form  k [ 1 - cos(theta) ], theta being the angle between the "z" axes of the two ligands,
        defined by the a2-a1 and b2-b1 distances, to keep the axes of the two ligands aligned.

        A potential of the form  k [ 1 - cos(psi) ], psi being the dihdral angle between the planes a3-a1-b2 and a1-b2-b3
        to keep the "roll" of the two ligands in alignment. This potential is symmetrized with respect to the
        exchange of ligand labels a and b.

        Parameters
        ----------
        liga_ref_particles : list of 3 ints
            List of the indexes of the reference atoms the first ligand
        ligb_ref_particles : list of 3 ints
            List of the indexes of the reference atoms of the second ligand
        kfdispl : Quantity in energy/distance^2 units
            Force constant of the quadratic potential
        ktheta : Quantity in energy units
            Height of the 1-cos(theta) restraint potential for the angle restraint
        kpsi : Quantity in energy units
            Height of the 1-cos(psi) restraint potential for the dihedral angle restraint
        offset : List of 3 Quantity's in distance units
            Center of the restraint. The potential is applied to the a1-b1 distance after subtracting offset.
            It is used to keep the a1-b1 distance near offset.

        Returns
        -------
        A tuple of OpenMM Forces with the displacement, the angle, and dihedral restraint.

        """

        if liga_ref_particles and ligb_ref_particles:
            if not (len(liga_ref_particles) == 3 and len(ligb_ref_particles) == 3):
                raise ValueError("Invalid lists of reference atoms")

        expr = " (kfdispl/2)*distsq  ; "  # displacement potential
        expr += " distsq = (x1 - offx - x2)^2 + "
        expr += "          (y1 - offy - y2)^2 + "
        expr += "          (z1 - offz - z2)^2   "  # square distance between b1 and a1 after displacing back b

        displforce = mm.CustomCompoundBondForce(2, expr)
        self.system.addForce(displforce)
        displforce.addPerBondParameter("kfdispl")
        displforce.addPerBondParameter("offx")
        displforce.addPerBondParameter("offy")
        displforce.addPerBondParameter("offz")
        offv = offset / nanometer
        offx = offv[0]
        offy = offv[1]
        offz = offv[2]
        displforce.addBond(
            [ligb_ref_particles[0], liga_ref_particles[0]],
            np.array(
                [kfdispl / ((kilojoule_per_mole / nanometer**2)), offx, offy, offz],
                dtype=np.double,
            ),
        )

        expr = "(ktheta/2) * (1 - cost) ;"  # theta restraint

        expr += "cost = xdn1*xdn2 + ydn1*ydn2 + zdn1*zdn2 ; "
        expr += "xdn1 = xd1/dn1 ; ydn1 = yd1/dn1 ; zdn1 = zd1/dn1 ;"
        expr += "dn1 = sqrt(xd1^2 + yd1^2 + zd1^2 ) ;"
        expr += "xd1 = x2 - x1 ; "
        expr += "yd1 = y2 - y1 ;"
        expr += "zd1 = z2 - z1 ;"
        expr += "xdn2 = xd2/dn2 ; ydn2 = yd2/dn2 ; zdn2 = zd2/dn2 ;"
        expr += "dn2 = sqrt(xd2^2 + yd2^2 + zd2^2 ) ;"
        expr += "xd2 = x4 - x3 ; "
        expr += "yd2 = y4 - y3 ; "
        expr += "zd2 = z4 - z3   "

        thetaforce = mm.CustomCompoundBondForce(4, expr)
        self.system.addForce(thetaforce)
        thetaforce.addPerBondParameter("ktheta")
        thetaforce.addBond(
            [
                ligb_ref_particles[0],
                ligb_ref_particles[1],
                liga_ref_particles[0],
                liga_ref_particles[1],
            ],
            np.array([ktheta / kilojoule_per_mole], dtype=np.double),
        )

        expr = "(kpsi/2) * (1 - cosp) ;"  # psi restraint

        expr += "cosp = xvn*xwn + yvn*ywn + zvn*zwn ; "
        expr += "xvn = xv/v ; yvn = yv/v; zvn = zv/v ;"
        expr += "v = sqrt(xv^2 + yv^2 + zv^2 ) ;"
        expr += "xv = xd0 - dot01*xdn1 ;"
        expr += "yv = yd0 - dot01*ydn1 ;"
        expr += "zv = zd0 - dot01*zdn1 ;"
        expr += "dot01 =  xd0*xdn1 +  yd0*ydn1 +  zd0*zdn1 ;"
        expr += "xd0 = x3 - x1 ;"
        expr += "yd0 = y3 - y1 ;"
        expr += "zd0 = z3 - z1 ;"
        expr += "xwn = xw/w ; ywn = yw/w; zwn = zw/w ;"
        expr += "w = sqrt(xw^2 + yw^2 + zw^2 ) ;"
        expr += "xw = xd3 - dot31*xdn1 ;"
        expr += "yw = yd3 - dot31*ydn1 ;"
        expr += "zw = zd3 - dot31*zdn1 ;"
        expr += "dot31 =  xd3*xdn1 +  yd3*ydn1 +  zd3*zdn1 ;"
        expr += "xd3 = x5 - x4 ;"
        expr += "yd3 = y5 - y4 ;"
        expr += "zd3 = z5 - z4 ; "
        expr += "xdn1 = xd1/dn1 ; ydn1 = yd1/dn1 ; zdn1 = zd1/dn1 ;"
        expr += "dn1 = sqrt(xd1^2 + yd1^2 + zd1^2 ) ;"
        expr += "xd1 = x2 - x1 ; "
        expr += "yd1 = y2 - y1 ;"
        expr += "zd1 = z2 - z1 "

        psiforce = mm.CustomCompoundBondForce(5, expr)
        self.system.addForce(psiforce)
        psiforce.addPerBondParameter("kpsi")
        psiforce.addBond(
            [
                ligb_ref_particles[0],
                ligb_ref_particles[1],
                ligb_ref_particles[2],
                liga_ref_particles[0],
                liga_ref_particles[2],
            ],
            np.array([0.5 * kpsi / kilojoule_per_mole], dtype=np.double),
        )
        psiforce.addBond(
            [
                liga_ref_particles[0],
                liga_ref_particles[1],
                liga_ref_particles[2],
                ligb_ref_particles[0],
                ligb_ref_particles[2],
            ],
            np.array([0.5 * kpsi / kilojoule_per_mole], dtype=np.double),
        )
        return (displforce, thetaforce, psiforce)

    def addVsiteRestraintForceCMAngles(
        self,
        lig_cm_groups=None,
        rcpt_cm_groups=None,
        ktheta=None,
        theta0=None,
        thetatol=None,
        kphi=None,
        phi0=None,
        phitol=None,
        kpsi=None,
        psi0=None,
        psitol=None,
    ):
        """Add orientation restraining potentials between ligand and receptor.

        It is used in conjunction with addVsiteRestraintForceCMCM() to
        define the spatial and angular extent of the receptor binding
        site.

        The orientation of the ligand with respect to the receptor is
        defined with respect to a Cartesian reference frame fixed on
        the receptor defined by 3, non co-linear, reference points
        (r1, r2, r3) determined by the centroids of 3 given groups of
        atoms of the receptor.  r1 defines the origin of the frame,
        the distance r2-r1 defines the z-axis of the frame, and the
        r3-r2-r1 plane defines the xz plane of the frame.  Similarly,
        the frame of the ligand is defined by 3 reference points (l1,
        l2, l3) determined by the centroids of 3 groups of atoms of
        the ligand molecule.

        The orientation angles defined below are those between the
        elements of the frame of the ligand with respect to the frame
        of the receptor when the two frames are shifted to a common
        origin such that r1 and l1 are both at (0,0,0).

        The theta angle is defined by the angle between the z-axes of
        the two frames, that is the angle between r2-r1 and l2-l1
        directions.

        The phi angle is the dihedral between the r3-r2-r1 plane and
        the r2-l1-l2 plane. That is, essentially, the azimuthal
        spherical polar angle of the l2-l1 ligand axis with respect to
        frame fixed on the receptor.

        The psi angle is the dihedral between the r2-r1-l2 plane and
        the l1-l2-l3 plane. That is, the torsional angle of the r2,
        l1=r1, l2, l3 points, or the "twist" around the z-axis of the
        ligand.

        The restraint potential added for each angle above (theta,
        phi, psi) is a flat-bottom quadratic potential defined by
        a center, force constant, and tolerance. The potential for
        theta is in terms of cos(theta). The potentials for phi
        and psi have a periodicity of 2pi.

        Parameters
        ----------
        lig_cm_groups: list of list of ints, such as [ [0,1], [2], [3] ]
            Each element of this list is the set of atom indexes of the ligand whose centroids
            define the reference points l1, l2, and l3, respectively.
        rcpt_cm_groups : list of list of ints, such as [ [4,5], [6], [7] ]
            Each element of this list is the set of atom indexes of the receptor whose centroids
            define the reference points r1, r2, and r3, respectively.
        ktheta : Quantity in energy units
            Force constant of the quadratic potential for cos(theta)
        theta0 : Quantity in angle units
            The cosine of this angle is the center of the flat-bottom potential for cos(theta)
        thetatol : Quantity in angle units
            Defines the tolerance of the flat-bottom potential for cos(theta) in terms
            of cos(theta0+-thetatol)
        kphi : Quantity in (energy/angle^2) units
            Force constant of the quadratic potential for the phi angle
        phi0 : Quantity in angle units
            The center of the flat-bottom potential for the phi angle
        phitol : Quantity in angle units
            Tolerance of the flat-bottom potential for the phi angle
        kpsi : Quantity in (energy/angle^2) units
            Force constant of the quadratic potential for the psi angle
        psi0 : Quantity in angle units
            The center of the flat-bottom potential for the psi angle
        psitol : Quantity in angle units
            Tolerance of the flat-bottom potential for the psi angle

        Returns
        -------
        A tuple of OpenMM Forces for the theta, phi, and psi restraints, respectively.

        """

        assert len(lig_cm_groups) == 3
        assert len(rcpt_cm_groups) == 3

        # theta restraint
        # (kf/2) (cos(theta) - cos(theta0))^2 with a tolerance
        thetaforce = None
        expr = "0.5*kf*( step(dcost)*max(0,dcost-ctol)^2 + step(-dcost)*max(0,-dcost-ctol)^2 ) ; dcost = cost - cos0 ; "
        expr += "cost = xdn1*xdn2 + ydn1*ydn2 + zdn1*zdn2 ; "
        expr += "xdn1 = xd1/dn1 ; ydn1 = yd1/dn1 ; zdn1 = zd1/dn1 ;"
        expr += "dn1 = sqrt(xd1^2 + yd1^2 + zd1^2 ) ;"
        expr += "xd1 = x2 - x1 ; "
        expr += "yd1 = y2 - y1 ;"
        expr += "zd1 = z2 - z1 ;"
        expr += "xdn2 = xd2/dn2 ; ydn2 = yd2/dn2 ; zdn2 = zd2/dn2 ;"
        expr += "dn2 = sqrt(xd2^2 + yd2^2 + zd2^2 ) ;"
        expr += "xd2 = x4 - x3 ; "
        expr += "yd2 = y4 - y3 ; "
        expr += "zd2 = z4 - z3   "
        if ktheta is not None:
            if self.CMAngleThetaForce is None:
                self.CMAngleThetaForce = mm.CustomCentroidBondForce(4, expr)
                thetaforce = self.CMAngleThetaForce
                numthetagroups = 0
                thetaforce.addPerBondParameter("kf")
                thetaforce.addPerBondParameter("cos0")
                thetaforce.addPerBondParameter("ctol")
                self.system.addForce(thetaforce)
            else:
                thetaforce = self.CMAngleThetaForce
                numthetagroups = thetaforce.getNumGroups()
            thetaforce.addGroup(rcpt_cm_groups[0])
            thetaforce.addGroup(rcpt_cm_groups[1])
            thetaforce.addGroup(lig_cm_groups[0])
            thetaforce.addGroup(lig_cm_groups[1])
            kf = ktheta / (kilojoule_per_mole)
            cos0 = math.cos(theta0 / radians)
            a0 = max(0, (theta0 - thetatol) / radians)
            b0 = min(math.pi, (theta0 + thetatol) / radians)
            ctol = abs(math.cos(a0) - math.cos(b0))
            groupids = [numthetagroups + i for i in range(4)]
            thetaforce.addBond(groupids, np.array([kf, cos0, ctol], dtype=np.double))

        # phi restraint
        # (kf/2) (phi - phi0)^2 with a tolerance and periodicity
        phiforce = None
        expr = "0.5*kf*( step(dm0)*max(0,db0)^2 + step(-dm0)*max(0,-da0)^2 ) ; "
        expr += self._wrapExpression("db0", "xb0", "twopi") + " ; xb0 = psi - b0 ; "
        expr += self._wrapExpression("da0", "xa0", "twopi") + " ; xa0 = psi - a0 ; "
        expr += (
            self._wrapExpression("dm0", "xm0", "twopi")
            + " ; xm0 = psi - mid0 ; mid0 = (a0 + b0)/2 ; "
        )
        expr += (
            self._dihedralExpression(
                "phi",
                "xr2",
                "yr2",
                "zr2",
                "xr1",
                "yr1",
                "zr1",
                "0",
                "0",
                "0",
                "xl1",
                "yl1",
                "zl1",
            )
            + " ; "
        )
        expr += (
            self._diffvExpression(
                "xr2", "yr2", "zr2", "x3", "y3", "z3", "x1", "y1", "z1"
            )
            + " ; "
        )
        expr += (
            self._diffvExpression(
                "xr1", "yr1", "zr1", "x2", "y2", "z2", "x1", "y1", "z1"
            )
            + " ; "
        )
        expr += (
            self._diffvExpression(
                "xl1", "yl1", "zl1", "x5", "y5", "z5", "x4", "y4", "z4"
            )
            + " ; "
        )
        expr += "twopi = 2*pi ;"
        expr += "pi = %f" % math.pi
        if kphi is not None:
            if self.CMAnglePhiForce is None:
                self.CMAnglePhiForce = mm.CustomCentroidBondForce(5, expr)
                phiforce = self.CMAnglePhiForce
                numphigroups = 0
                phiforce.addPerBondParameter("kf")
                phiforce.addPerBondParameter("a0")
                phiforce.addPerBondParameter("b0")
                self.system.addForce(phiforce)
            else:
                phiforce = self.CMAnglePhiForce
                numphigroups = phiforce.getNumGroups()
            phiforce.addGroup(rcpt_cm_groups[0])  # 1
            phiforce.addGroup(rcpt_cm_groups[1])  # 2
            phiforce.addGroup(rcpt_cm_groups[2])  # 3
            phiforce.addGroup(lig_cm_groups[0])  # 4
            phiforce.addGroup(lig_cm_groups[1])  # 5
            kf = kphi / (kilojoule_per_mole / radians**2)
            a0 = (phi0 - phitol) / radians
            b0 = (phi0 + phitol) / radians
            groupids = [numphigroups + i for i in range(5)]
            phiforce.addBond(groupids, np.array([kf, a0, b0], dtype=np.double))

        # psi restraint
        # (kf/2) (psi - psi0)^2 with a tolerance and periodicity
        psiforce = None
        expr = "0.5*kf*( step(dm0)*max(0,db0)^2 + step(-dm0)*max(0,-da0)^2 ) ; "
        expr += self._wrapExpression("db0", "xb0", "twopi") + " ; xb0 = psi - b0 ; "
        expr += self._wrapExpression("da0", "xa0", "twopi") + " ; xa0 = psi - a0 ; "
        expr += (
            self._wrapExpression("dm0", "xm0", "twopi")
            + " ; xm0 = psi - mid0 ; mid0 = (a0 + b0)/2 ; "
        )
        expr += (
            self._dihedralExpression(
                "psi",
                "xr1",
                "yr1",
                "zr1",
                "0",
                "0",
                "0",
                "xl1",
                "yl1",
                "zl1",
                "xl2",
                "yl2",
                "zl2",
            )
            + " ; "
        )
        expr += (
            self._diffvExpression(
                "xr1", "yr1", "zr1", "x2", "y2", "z2", "x1", "y1", "z1"
            )
            + " ; "
        )
        expr += (
            self._diffvExpression(
                "xl1", "yl1", "zl1", "x4", "y4", "z4", "x3", "y3", "z3"
            )
            + " ; "
        )
        expr += (
            self._diffvExpression(
                "xl2", "yl2", "zl2", "x5", "y5", "z5", "x3", "y3", "z3"
            )
            + " ; "
        )
        expr += "twopi = 2*pi ;"
        expr += "pi = %f" % math.pi
        if kpsi is not None:
            if self.CMAnglePsiForce is None:
                self.CMAnglePsiForce = mm.CustomCentroidBondForce(5, expr)
                psiforce = self.CMAnglePsiForce
                numpsigroups = 0
                psiforce.addPerBondParameter("kf")
                psiforce.addPerBondParameter("a0")
                psiforce.addPerBondParameter("b0")
                self.system.addForce(psiforce)
            else:
                psiforce = self.CMAnglePsiForce
                numpsigroups = psiforce.getNumGroups()
            psiforce.addGroup(rcpt_cm_groups[0])  # 1
            psiforce.addGroup(rcpt_cm_groups[1])  # 2
            psiforce.addGroup(lig_cm_groups[0])  # 3
            psiforce.addGroup(lig_cm_groups[1])  # 4
            psiforce.addGroup(lig_cm_groups[2])  # 5
            kf = kpsi / (kilojoule_per_mole / radians**2)
            a0 = (psi0 - psitol) / radians
            b0 = (psi0 + psitol) / radians
            groupids = [numpsigroups + i for i in range(5)]
            psiforce.addBond(groupids, np.array([kf, a0, b0], dtype=np.double))
        return (thetaforce, phiforce, psiforce)

    def addPosRestraints(
        self,
        particles,
        refpos,
        fc=25.0 * kilocalorie_per_mole / angstrom**2,
        tol=0.5 * angstrom,
        periodic=True,
    ):
        """
        Applies flat-bottom, quadratic position restraints to a set of
        atoms based on a set of reference positions.

        Parameters
        ----------
        particles: list of ints
            List of the indexes of the atoms to restrain
        refpos : list of Vec3
            Centers of the flat-bottom restraints. These are reference
            positions for all atoms of the system,
            not just those that are restrained.
        fc : Quantitiy in (energy/distance^2) units
            Force constant of the quadratic flat-bottom potential
        tol : Quantity in distance units
            Tolerance of the quadratic flat-bottom potential

        Returns
        -------
        The Force added to the System

        """

        if not particles or len(particles) == 0:
            return None
        if periodic:
            posrestforce = mm.CustomExternalForce(
                "0.5*fc*select(step(dist-tol), (dist-tol)^2, 0); dist = periodicdistance(x,y,z,x0,y0,z0)"
            )
        else:
            posrestforce = mm.CustomExternalForce(
                "0.5*fc*select(step(dist-tol), (dist-tol)^2, 0); dist = sqrt((x-x0)^2+(y-y0)^2+(z-z0)^2)"
            )

        posrestforce.addPerParticleParameter("x0")
        posrestforce.addPerParticleParameter("y0")
        posrestforce.addPerParticleParameter("z0")
        posrestforce.addPerParticleParameter("fc")
        posrestforce.addPerParticleParameter("tol")

        self.system.addForce(posrestforce)

        for p in particles:
            x0 = refpos[p][0] / nanometer
            y0 = refpos[p][1] / nanometer
            z0 = refpos[p][2] / nanometer
            fc1 = fc / (kilojoule_per_mole / nanometer**2)
            tol1 = tol / nanometer
            posrestforce.addParticle(
                p, np.array([x0, y0, z0, fc1, tol1], dtype=np.double)
            )
        return posrestforce

    # fix zero LJs
    def fixZeroLJParams(
        self, force, minsigma=0.1 * angstrom, minepsilon=1.0e-4 * kilocalories_per_mole
    ):
        """
        Set the Lennard-Jones parameters to some minimum values

        This is used for unphysical force fields that do not have
        a ground state because of "naked" charge centers that can
        sit on other charge centers without penalty.

        Parameters
        ----------
        force : OpenMM force
            Non-bonded Force to operate upon, other Forces are skipped
        minsigma : Quantity in distance units
            Minimum value of the sigma LJ parameter to assign
        minepsilone : Quantity in energy units
            Minimum value of the epsilon LJ parameter to assign

        """

        small = 1.0e-6
        if isinstance(force, mm.NonbondedForce):
            for i in range(force.getNumParticles()):
                nbparam = force.getParticleParameters(i)
                charge = nbparam[0]
                sigmaLJ = nbparam[1]
                epsilonLJ = nbparam[2]
                if sigmaLJ._value < small and epsilonLJ._value < small:
                    sigmaLJ = minsigma
                    epsilonLJ = minepsilon
                    force.setParticleParameters(i, charge, sigmaLJ, epsilonLJ)

    def softCorePertE(self, u, umax, ub, a):
        usc = u
        if u > ub:
            gu = (u - ub) / (a * (umax - ub))  # this is y/alpha
            zeta = 1.0 + 2.0 * gu * (gu + 1.0)
            zetap = np.power(zeta, a)
            usc = (umax - ub) * (zetap - 1.0) / (zetap + 1.0) + ub
        return usc



import os
import logging.config
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("func2argparse")
except PackageNotFoundError:
    pass


dirname = os.path.dirname(__file__)
try:
    logging.config.fileConfig(
        os.path.join(dirname, "logging.ini"), disable_existing_loggers=False
    )
except Exception:
    print("atm: Logging setup failed")



import logging
import math
from openmm.unit import kelvin, kilocalories_per_mole
from atm.utils import pairwise_independence_sampling
from atm.ommreplica import OMMReplicaATM
from atm.ommsystem import OMMSystemAmberRBFE
from atm.worker import OMMWorkerATM
from atm.utils import TerminationGuard, Timer
import sys

logger = logging.getLogger(__name__)


class OpenmmJobAmberRBFE:

    def __init__(self, config_file):
        import yaml
        import json

        self.logger = logger

        self.logger.info("Configuration:")
        if config_file.endswith(".yaml"):
            with open(config_file, "r") as f:
                self.config = yaml.safe_load(f)
        elif config_file.endswith(".json"):
            with open(config_file, "r") as f:
                self.config = json.load(f)
        else:
            raise ValueError("Invalid configuration file format")

        for key, value in self.config.items():
            self.logger.info(f"{key}: {value}")

        if self.config.get("VERBOSE"):
            self.logger.setLevel(logging.DEBUG)

        self.basename = self.config["BASENAME"]
        self.state_params = self._getStateParams()
        self.nreplicas = len(self.state_params)

    def _getStateParams(self):
        lambdas = self.config["LAMBDAS"]
        directions = self.config["DIRECTION"]
        intermediates = self.config["INTERMEDIATE"]
        lambda1s = self.config["LAMBDA1"]
        lambda2s = self.config["LAMBDA2"]
        alphas = self.config["ALPHA"]
        uhs = self.config["U0"]
        w0s = self.config["W0COEFF"]
        temperatures = self.config["TEMPERATURES"]

        assert len(directions) == len(lambdas)
        assert len(intermediates) == len(lambdas)
        assert len(lambda1s) == len(lambdas)
        assert len(lambda2s) == len(lambdas)
        assert len(alphas) == len(lambdas)
        assert len(uhs) == len(lambdas)
        assert len(w0s) == len(lambdas)
        assert len(temperatures) == 1

        self.logger.info("State parameters")
        state_params = []
        for lambda_, direction, intermediate, lambda1, lambda2, alpha, uh, w0 in zip(
            lambdas, directions, intermediates, lambda1s, lambda2s, alphas, uhs, w0s
        ):
            par = {}
            par["lambda"] = float(lambda_)
            par["atmdirection"] = float(direction)
            par["atmintermediate"] = float(intermediate)
            par["lambda1"] = float(lambda1)
            par["lambda2"] = float(lambda2)
            par["alpha"] = float(alpha) / kilocalories_per_mole
            par["uh"] = float(uh) * kilocalories_per_mole
            par["Umax"] = float(self.config.get("UMAX")) * kilocalories_per_mole
            par["Ubcore"] = float(self.config.get("UBCORE")) * kilocalories_per_mole
            par["Acore"] = float(self.config.get("ACORE"))
            par["w0"] = float(w0) * kilocalories_per_mole
            par["temperature"] = float(temperatures[0]) * kelvin
            state_params.append(par)
            self.logger.info(f"    State: {par}")

        return state_params

    def setupJob(self):
        with Timer(self.logger.info, "ATM setup"):

            with Timer(self.logger.info, "create system"):
                prmtopfile = self.basename + ".prmtop"
                crdfile = self.basename + ".inpcrd"
                ommsystem = OMMSystemAmberRBFE(
                    self.basename, self.config, prmtopfile, crdfile, self.logger
                )
                ommsystem.create_system()

            with Timer(self.logger.info, "create worker"):
                self.worker = OMMWorkerATM(ommsystem, self.config, self.logger)

            with Timer(self.logger.info, "create replicas"):
                self.replicas = []
                for i in range(self.nreplicas):
                    replica = OMMReplicaATM(i, self.basename, self.worker, self.logger)
                    if not replica.get_stateid():
                        replica.set_state(i, self.state_params[i])
                    self.replicas.append(replica)

                self.replica_states = [
                    replica.get_stateid() for replica in self.replicas
                ]
                for i, replica in enumerate(self.replicas):
                    self.logger.info(
                        f"Replica {i}: cycle {replica.get_cycle()}, state {replica.get_stateid()}"
                    )

            with Timer(self.logger.info, "update replicas"):
                self._updateReplicas()

    def scheduleJobs(self):
        import os

        with Timer(self.logger.info, "ATM simulations"):
            last_sample = self.replicas[0].get_cycle()
            num_samples = self.config["MAX_SAMPLES"]

            write_progress = False
            starting_sample = 0
            if isinstance(num_samples, str) and num_samples.startswith("+"):
                # Handle cases where we want to increase the number of samples from a starting checkpoint
                write_progress = True
                num_samples = int(num_samples[1:])
                if not os.path.isfile("starting_sample"):
                    with open("starting_sample", "w") as f:
                        f.write(f"{last_sample}\n")
                with open("starting_sample", "r") as f:
                    starting_sample = int(f.read().strip())
                    last_sample = (last_sample - starting_sample) + 1

            self.logger.info(f"Target number of samples: {num_samples}")

            # in memory storage for output data
            output_data = [[] for _ in range(len(self.replicas))]

            for isample in range(last_sample, num_samples + 1):
                if write_progress:
                    with open("progress", "w") as f:
                        f.write(f"{(isample - 1) / num_samples}\n")

                with Timer(self.logger.info, f"sample {isample}"):

                    for irepl, replica in enumerate(self.replicas):
                        with Timer(
                            self.logger.info, f"sample {isample}, replica {irepl}"
                        ):
                            if starting_sample == 0:
                                assert replica.get_cycle() == isample
                            self.worker.run(replica)

                    with Timer(self.logger.info, "exchange replicas"):
                        self._exhangeReplicas()

                    with Timer(self.logger.info, "update replicas"):
                        self._updateReplicas()

                    with Timer(self.logger.info, "save replica samples"):
                        for irepl, replica in enumerate(self.replicas):
                            # save output data to write at same freq at checkpoint
                            data = replica.save_out()
                            output_data[irepl].append(data)

                    if (
                        isample % int(self.config["CHECKPOINT_FREQUENCY"]) == 0
                        or isample == num_samples
                    ):
                        # at every checkpoint_frequency samples write the the .out files and the checkpoint xml.
                        # also write the output if this is the last sample

                        with TerminationGuard():
                            with Timer(
                                self.logger.info,
                                "write replicas samples and trajectories",
                            ):

                                if not all(
                                    len(lst) == len(output_data[0])
                                    for lst in output_data
                                ):
                                    self.logger.error(
                                        "Inconsistent output data. Did one of the replicas crash?"
                                    )
                                    raise ValueError

                                for irepl, replica in enumerate(self.replicas):
                                    # write multiple lines to the out file
                                    replica.write_out(output_data[irepl])
                                    # clear the saved lines
                                    output_data[irepl] = []

                                    if (
                                        replica.get_mdsteps()
                                        % int(self.config["TRJ_FREQUENCY"])
                                        == 0
                                    ):
                                        replica.save_dcd()

                            with Timer(self.logger.info, "checkpointing"):
                                for replica in self.replicas:
                                    replica.save_checkpoint()

    def _updateReplicas(self):
        for replica, stateid in zip(self.replicas, self.replica_states):
            replica.set_state(stateid, self.state_params[stateid])

    def _exhangeReplicas(self):

        # Matrix of replica energies in each state.
        swap_matrix = self._computeSwapMatrix(
            range(self.nreplicas), self.replica_states
        )
        self.logger.debug("Swap matrix")
        for row in swap_matrix:
            self.logger.debug(f"    {row}")

        self.logger.debug(f"Replica states before: {self.replica_states}")
        for repl_i in range(self.nreplicas):
            sid_i = self.replica_states[repl_i]
            repl_j = pairwise_independence_sampling(
                repl_i, sid_i, range(self.nreplicas), self.replica_states, swap_matrix
            )
            if repl_j != repl_i:
                sid_i = self.replica_states[repl_i]
                sid_j = self.replica_states[repl_j]
                self.replica_states[repl_i] = sid_j
                self.replica_states[repl_j] = sid_i
                self.logger.info(f"Replica {repl_i}: {sid_i} --> {sid_j}")
                self.logger.info(f"Replica {repl_j}: {sid_j} --> {sid_i}")

        self.logger.debug(f"Replica states after: {self.replica_states}")

    def _computeSwapMatrix(self, repls, states):
        """
        Compute matrix of dimension-less energies: each column is a replica
        and each row is a state so U[i][j] is the energy of replica j in state
        i.
        """
        U = [[0.0 for _ in range(self.nreplicas)] for _ in range(self.nreplicas)]

        n = len(repls)

        # collect replica parameters and potentials
        par = [self._getPar(k) for k in repls]
        pot = [self._getPot(k) for k in repls]

        for i in range(n):
            repl_i = repls[i]
            for j in range(n):
                sid_j = states[j]
                # energy of replica i in state j
                U[sid_j][repl_i] = _reduced_energy(par[j], pot[i])
        return U

    def _getPar(self, repl):
        _, par = self.replicas[repl].get_state()
        return par

    # customized getPot to return the unperturbed potential energy
    # of the replica U0 = U - W_lambda(u)
    def _getPot(self, repl):
        replica = self.replicas[repl]
        pot = replica.get_energy()
        epot = pot["potential_energy"]
        pertpot = pot["perturbation_energy"]
        _, par = replica.get_state()
        lambda1 = par["lambda1"]
        lambda2 = par["lambda2"]
        alpha = par["alpha"]
        uh = par["uh"]
        w0 = par["w0"]
        ebias = _softplus(lambda1, lambda2, alpha, uh, w0, pertpot)
        pot["unbiased_potential_energy"] = epot - ebias
        pot["direction"] = par["atmdirection"]
        pot["intermediate"] = par["atmintermediate"]
        return pot


def _reduced_energy(par, pot):
    lambda1 = par["lambda1"]
    lambda2 = par["lambda2"]
    alpha = par["alpha"]
    uh = par["uh"]
    w0 = par["w0"]
    state_direction = par["atmdirection"]
    state_intermediate = par["atmintermediate"]
    epot0 = pot["unbiased_potential_energy"]
    pertpot = pot["perturbation_energy"]
    replica_direction = pot["direction"]
    replica_intermediate = pot["intermediate"]

    kb = 0.0019872041 * kilocalories_per_mole / kelvin
    beta = 1.0 / (kb * par["temperature"])

    if (replica_direction == state_direction) or (
        state_intermediate > 0 and replica_intermediate > 0
    ):
        ebias = _softplus(lambda1, lambda2, alpha, uh, w0, pertpot)
        return beta * (epot0 + ebias)
    else:
        # prevent exchange
        large_energy = 1.0e12
        return large_energy


# evaluates the softplus function
def _softplus(lambda1, lambda2, alpha, uh, w0, uf):
    ee = 1.0 + math.exp(-alpha * (uf - uh))
    softplusf = lambda2 * uf + w0
    if alpha._value > 0.0:
        softplusf += ((lambda2 - lambda1) / alpha) * math.log(ee)
    return softplusf


def rbfe_production(config_file=None):
    from atm.utils import set_directory
    from pathlib import Path
    import os

    if config_file is None:
        config_file = sys.argv[1]

    with set_directory(Path(config_file).parent):
        rx = OpenmmJobAmberRBFE(os.path.basename(os.path.abspath(config_file)))
        rx.setupJob()
        rx.scheduleJobs()


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Specify ONE input file"

    rbfe_production(sys.argv[1])



from contextlib import contextmanager
import time
from signal import signal, SIGINT, SIGTERM
from pathlib import Path
import sys
import os
import warnings


class TerminationGuard:
    def __enter__(self):
        self.terminate = False
        self.sigint = signal(SIGINT, self)
        self.sigterm = signal(SIGTERM, self)

    def __call__(self, *_):
        self.terminate = True
        warnings.warn(
            "The process termination is delayed until a critical operation is completed"
        )

    def __exit__(self, *_):
        signal(SIGINT, self.sigint)
        signal(SIGTERM, self.sigterm)
        if self.terminate:
            sys.exit(0)


@contextmanager
def Timer(logger, message):
    logger(f"Started: {message}")
    start = time.monotonic()
    yield
    duration = time.monotonic() - start
    logger(f"Finished: {message} (duration: {duration} s)")


def _exit(message):
    """Print and flush a message to stdout and then exit."""
    print(message)
    sys.stdout.flush()
    print("exiting...")
    sys.exit(1)


def weighted_choice(choices):
    import numpy as np

    """Return a discrete outcome given a set of outcome/weight pairs."""
    r = np.random.random() * np.sum(w for c, w in list(choices))
    for c, w in choices:
        r -= w
        if r < 0:
            return c
    # You should never get here.
    return None


def pairwise_independence_sampling(repl_i, sid_i, replicas, states, U):
    """
    Return a replica "j" to exchange with the given replica "i" based on
    independent sampling from the discrete Metropolis transition matrix, T:

    T_rs = alpha_rs min[1,exp(-du_rs)]  r != s
    T_rr = 1 - sum_(s != r) T_rs        otherwise

    Here r and s are state index permutations, r being the current permutation
    and s the new permutation. alpha_rs = 0 unless r and s differ by a single
    replica swap and alpha_rs = 1/(n-1) otherwise, n being the number of
    replicas/states and (n-1) is the number of permutations, s, differing by
    permutation r by a single swap. du_rs is the change in reduced potential
    energy of the replica exchange ensemble in going from permutation r to
    permutation s (that is, due to a replica swap). Based on the above we have:

    du_rs = u_a(j)+u_b(i)-[u_a(i)+u_b(j)]

    where i and j are the replicas being swapped and a and b are, respectively,
    the states they occupy in r and b and a, respectively, those in s.

    The energies u_a(i), i=1,n and a=1,n, are assumed stored in the input
    "swap matrix," U[a][i].

    In general, the set of replicas across which exchanges are considered is a
    subset of the n replicas. This list is passed in the 'replicas' list.
    Replica "i" ('repl_i') is assumed to be in this list.
    """
    # Evaluate all i-j swap probabilities.
    #
    import numpy as np

    nreplicas = len(replicas)
    ps = np.zeros(nreplicas)  # probability of swap i <-> j
    du = np.zeros(nreplicas)  # Boltzmann exponent, ps ~ exp(-du)

    for j, repl_j, sid_j in zip(range(nreplicas), replicas, states):
        du[j] = (
            U[sid_i][repl_j] + U[sid_j][repl_i] - U[sid_i][repl_i] - U[sid_j][repl_j]
        )
    eu = np.exp(-du)

    pii = 1.0
    i = -1
    f = 1.0 / (float(nreplicas) - 1.0)
    for j in range(nreplicas):
        repl_j = replicas[j]
        if repl_j == repl_i:
            i = j
        else:
            if eu[j] > 1.0:
                ps[j] = f
            else:
                ps[j] = f * eu[j]
            pii -= ps[j]
    try:
        ps[i] = pii
    except IndexError:
        _exit(
            "gibbs_re_j(): unrecoverable error: replica %d not in the "
            "list of waiting replicas?" % i
        )

    return replicas[weighted_choice(list(zip(range(nreplicas), ps)))]


@contextmanager
def set_directory(path: Path):
    """Sets the cwd within the context

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """

    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)



import shutil
import os
import pytest


curr_dir = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize("jobname", ["QB_A08_A07", "QB_A15_A14", "QB_A16_A06"])
def _test_regression(tmp_path, jobname):
    from atm.rbfe_structprep import rbfe_structprep
    from atm.rbfe_production import rbfe_production
    from atm.uwham import calculate_uwham

    exp_ddG = {
        "QB_A08_A07": 0.78,
        "QB_A15_A14": 1.28,
        "QB_A16_A06": -1.23,
    }[jobname]
    expected_error = {
        "QB_A08_A07": 0.50,
        "QB_A15_A14": 0.35,
        "QB_A16_A06": 0.35,
    }[jobname]

    shutil.copytree(os.path.join(curr_dir, jobname), os.path.join(tmp_path, jobname))
    rbfe_structprep(os.path.join(tmp_path, jobname, f"{jobname}_asyncre.yaml"))
    rbfe_production(os.path.join(tmp_path, jobname, f"{jobname}_asyncre.yaml"))

    ddG = calculate_uwham(os.path.join(tmp_path, jobname), jobname, 100)[0]

    assert (
        abs(ddG - exp_ddG) < expected_error
    ), f"Predicted ddG: {ddG}, expected value: {exp_ddG}, acceptable error: {expected_error}"



import numpy as np
from scipy.optimize import minimize


def insert(x, d, x0=0):
    """Insert a value x0 at d-th position of x"""
    if d == 0:
        return np.concatenate(([x0], x))
    else:
        return np.concatenate((x[:d], [x0], x[d:]))


def obj_fcn(ze, logQ, size, base):
    """
    Objective function for UWHAM optimization

    Args:
        ze: vector of log-normalizing constants (or free energies)
        logQ: log of unnormalized density ratios (over baseline)
        size: individual sample sizes for distributions
        base: baseline index
    """
    N = logQ.shape[1]
    rho = size / N

    ze = insert(ze, base, 0)

    # Calculate normalized Q values
    Qnorm = np.exp(logQ - ze[:, np.newaxis]) * rho[:, np.newaxis]
    Qsum = np.sum(Qnorm, axis=0)

    val = np.sum(np.log(Qsum)) / N + np.sum(ze * rho)

    # Calculate gradient and hessian
    W = np.delete(Qnorm, base, axis=0).T / Qsum[:, None]
    grad = -np.sum(W, axis=0) / N + np.delete(rho, base)

    O = W.T @ W / N
    hess = -O + np.diag(np.sum(W, axis=0)) / N

    return val, grad, hess


def uwham(logQ, size=None, label=None, base=None, init=None, fisher=True):
    """
    Unbinned Weighted Histogram Analysis Method

    Args:
        logQ: N x M matrix of log unnormalized densities
        size: vector of length M giving sample sizes for distributions
        label: vector of labels indicating which observation is from which distribution
        base: baseline index (0 to M-1) for distribution whose free energy is set to 0
        init: initial values of free energies
        fisher: if True, variance estimation based on Fisher information
    """
    N, M = logQ.shape

    # Compute size if needed
    if size is None:
        if label is None:
            raise ValueError("Either label or size must be provided")
        size = np.bincount(label, minlength=M)[1:]

    # Check which distributions have samples
    sampled = size > 0

    # Validate inputs
    if N != np.sum(size):
        raise ValueError("Inconsistent sum(size) and logQ.shape[0]")
    if M != len(size):
        raise ValueError("Inconsistent length(size) and logQ.shape[1]")

    # Check size and label consistency
    if label is not None:
        _, counts = np.unique(label[label >= 0], return_counts=True)
        if not np.array_equal(counts, size[sampled]):
            raise ValueError("Inconsistent label and size[sampled]")
    else:
        if fisher is False:
            label = np.repeat(np.where(sampled)[0], size[sampled])
            print(
                "Assume observations are ordered by thermodynamic state if fisher=False and label=None"
            )

    # Set base if not provided
    if base is None:
        base = np.where(sampled)[0][0]
    elif not sampled[base]:
        raise ValueError("Observations from baseline are required")

    # Set initial values if not provided
    if init is None:
        init = np.zeros(M)

    # Get number of sampled distributions
    m = np.sum(sampled)
    rho = size[sampled] / N

    # Calculate baseline index for sampled distributions
    temp = np.zeros(M - 1)
    temp = np.insert(temp, base, 1)
    base0 = np.where(temp[sampled])[0][0]

    # Calculate log density ratios
    logQ = logQ - logQ[:, base : base + 1]
    logQ = logQ.T

    # Optimize using scipy
    result = minimize(
        lambda x: obj_fcn(x, logQ[sampled], size[sampled], base0)[0],
        init[sampled][np.arange(len(init[sampled])) != base0],
        method="trust-exact",
        jac=lambda x: obj_fcn(x, logQ[sampled], size[sampled], base0)[1],
        hess=lambda x: obj_fcn(x, logQ[sampled], size[sampled], base0)[2],
    )

    # Process optimization results
    ze0 = insert(result.x, base0)
    ze = np.zeros(M)
    ze[sampled] = ze0

    # Calculate normalized weights
    Qnorm = np.exp(logQ - ze[:, None])
    Qsum = np.sum(Qnorm[sampled] * rho[:, None], axis=0)

    W = Qnorm.T / Qsum[:, None]
    z = np.mean(W, axis=0)

    # Check values (should all be 1)
    check = z[sampled]

    # Handle unsampled distributions
    if m < M:
        ze[~sampled] = np.log(z[~sampled])
        W[:, ~sampled] = W[:, ~sampled] / z[~sampled]

    # Return early if no variance estimation needed
    if fisher is None:
        return {
            "ze": ze,
            "W": W,
            "check": check,
            "result": result,
            "size": size,
            "base": base,
        }

    # Variance estimation
    O = W.T @ W / N

    D = np.zeros((M, M))
    D[:, sampled] = O[:, sampled] * rho

    H = D - np.eye(M)
    H = np.delete(np.delete(H, base, 0), base, 1)

    if fisher:
        G = O - D[:, sampled] @ O[sampled]
        G = np.delete(np.delete(G, base, 0), base, 1)

        iHG = -O + np.ones((M, 1)) @ O[base : base + 1]
        iHG = np.delete(np.delete(iHG, base, 0), base, 1)

        Ve = iHG @ np.linalg.solve(H.T, np.eye(H.shape[0])) / N

    else:
        C = np.zeros((m, M))
        for j in range(M):
            C[:, j] = np.bincount(label, weights=W[:, j], minlength=M)[sampled]

        G = O - C.T @ np.diag(rho) @ C
        G = np.delete(np.delete(G, base, 0), base, 1)
        Ve = np.linalg.solve(H, G) @ np.linalg.solve(H.T, np.eye(H.shape[0])) / N

    # Variance vector
    ve = insert(np.diag(Ve), base)

    # Variance-covariance matrix
    Ve = np.insert(Ve, base, 0, axis=1)
    Ve = np.insert(Ve, base, 0, axis=0)

    return {
        "ze": ze,
        "ve": ve,
        "Ve": Ve,
        "W": W,
        "check": check,
        "result": result,
        "label": label,
        "size": size,
        "base": base,
    }


def bias_fcn(epert, lam1, lam2, alpha, u0, w0):
    """
    Bias ilogistic potential:
    (lambda2-lambda1) ln[1+exp(-alpha (u-u0))]/alpha + lambda2 u + w0
    """
    ebias1 = np.zeros_like(epert)
    if alpha > 0:
        ee = 1 + np.exp(-alpha * (epert - u0))
        ebias1 = (lam2 - lam1) * np.log(ee) / alpha
    return ebias1 + lam2 * epert + w0


def dbias_fcn(epert, lam1, lam2, alpha, u0, w0):
    """
    Derivative of the bias ilogistic potential:
    (lambda2-lambda1)/(1+exp(-alpha (u-u0))) + lambda2
    """
    ee = 1.0 + np.exp(-alpha * (epert - u0))
    return (lam2 - lam1) / ee + lam1


def npot_fcn(e0, epert, bet, lam1, lam2, alpha, u0, w0):
    """
    Negative reduced energy: -beta*(U0+bias)
    """
    return -bet * (e0 + bias_fcn(epert, lam1, lam2, alpha, u0, w0))


def uwham_r(label, logQ, ufactormax, ufactormin=1):
    """
    UWHAM implementation
    Note: This is a simplified version and would need a proper UWHAM implementation
    """
    n = logQ.shape[0]
    m = logQ.shape[1]
    iniz = np.zeros(m)
    uf = ufactormax

    while uf >= ufactormin and uf >= 1:
        mask = slice(0, n, int(uf))
        out = uwham(label=label[mask], logQ=logQ[mask], init=iniz)
        iniz = out["ze"]
        uf = uf / 2

    return out


def calculate_uwham(
    rundir,
    jobname,
    mintimeid=None,
    maxtimeid=None,
    intermd=None,
    lambda1=None,
    lambda2=None,
    alpha=None,
    u0=None,
    w0=None,
):
    import pandas as pd
    import os

    # Define states
    tempt = np.array([300])
    bet = 1.0 / (0.001986209 * tempt)
    # fmt: off
    # directn = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])
    if intermd is None:
        intermd = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    else:
        intermd = np.array(intermd)
    nstates = len(intermd)
    if lambda1 is None:
        lambda1 = np.array([0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.10, 0.20, 0.30, 0.40, 0.50, 0.50,
                            0.40, 0.30, 0.20, 0.10, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00])
    else:
        lambda1 = np.array(lambda1)
    if lambda2 is None:
        lambda2 = np.array([0.00, 0.10, 0.20, 0.30, 0.40, 0.50, 0.50, 0.50, 0.50, 0.50, 0.50, 0.50,
                            0.50, 0.50, 0.50, 0.50, 0.50, 0.40, 0.30, 0.20, 0.10, 0.00])
    else:
        lambda2 = np.array(lambda2)
    if alpha is None:
        alpha = np.full(nstates, 0.10)
    else:
        alpha = np.array(alpha)
    if u0 is None:
        u0 = np.full(nstates, 110.0)
    else:
        u0 = np.array(u0)
    if w0 is None:
        w0 = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    else:
        w0 = np.array(w0)
    # fmt: on

    # Calculate states
    # np.where returns a tuple, we want the first element, then take the first occurrence
    leg1istate = np.where(intermd == 1)[0][0]
    leg2istate = np.where(intermd == 1)[0][1]

    # Define column names
    columns = [
        "stateid",
        "temperature",
        "direction",
        "lambda1",
        "lambda2",
        "alpha",
        "u0",
        "w0",
        "potE",
        "pertE",
        "trash",
    ]

    # Create list of data files
    datafiles = [
        os.path.join(rundir, f"r{i}", f"{jobname}.out") for i in range(nstates)
    ]

    # Read and combine all data files
    dfs = []
    for i, file in enumerate(datafiles):
        # Read data file
        df = pd.read_csv(file, sep=r"\s+", header=None, names=columns, index_col=False)
        # Add timeid column
        df["timeid"] = np.arange(1, len(df) + 1)
        dfs.append(df)

    # Combine all dataframes
    data = pd.concat(dfs, ignore_index=True)

    # Add beta column
    data["bet"] = 1.0 / (0.001986209 * data["temperature"])

    # Calculate time masks
    timemask = np.ones(len(data), dtype=bool)
    if mintimeid is not None:
        timemask = timemask & (data["timeid"] >= mintimeid)
    if maxtimeid is not None:
        timemask = timemask & (data["timeid"] <= maxtimeid)

    # Calculate samples
    nsamples = len(data[timemask])
    samplesperreplica = nsamples // nstates

    # Filter data for leg1
    data1 = data[timemask & (data["stateid"] <= leg1istate)]

    mtempt = len(bet)
    leg1stateids = np.arange(leg1istate + 1)
    mlam = len(leg1stateids)
    m = mlam * mtempt
    N = len(data1)

    # Extract U0 values as U-bias
    e0 = data1["potE"].values.copy()
    for i in range(N):
        e0[i] -= bias_fcn(
            data1["pertE"].iloc[i],
            data1["lambda1"].iloc[i],
            data1["lambda2"].iloc[i],
            data1["alpha"].iloc[i],
            data1["u0"].iloc[i],
            data1["w0"].iloc[i],
        )

    # Calculate negative potential
    neg_pot = np.zeros((N, m))
    sid = 0
    for be in leg1stateids:
        for te in range(mtempt):
            neg_pot[:, sid] = npot_fcn(
                e0,
                data1["pertE"].values,
                bet[te],
                lambda1[be],
                lambda2[be],
                alpha[be],
                u0[be],
                w0[be],
            )
            sid += 1

    # the alchemical state indexes start with 0, UWHAM's state labels start with 1
    statelabels = data1["stateid"].values.astype(int) + 1

    # Run UWHAM (placeholder - needs implementation)
    out = uwham_r(label=statelabels, logQ=neg_pot, ufactormax=1, ufactormin=1)

    # Reshape results
    ze = np.array(out["ze"]).reshape(mtempt, mlam)
    ve = np.array(out["ve"]).reshape(mtempt, mlam)

    # Calculate binding free energies
    dgbind1 = (-ze[:, -1] / bet) - (-ze[:, 0] / bet)
    ddgbind1 = np.sqrt(ve[:, -1] + ve[:, 0]) / bet

    # Filter data for leg 2
    data1 = data[timemask & (data["stateid"] >= leg2istate)]

    # Create reversed sequence for leg2 states
    leg2stateids = np.arange(leg2istate, nstates)[::-1]

    mlam = len(leg2stateids)
    m = mlam * mtempt
    N = len(data1)

    # Extract U0 values as U-bias
    e0 = data1["potE"].copy()
    for i in range(N):
        e0.iloc[i] -= bias_fcn(
            data1["pertE"].iloc[i],
            data1["lambda1"].iloc[i],
            data1["lambda2"].iloc[i],
            data1["alpha"].iloc[i],
            data1["u0"].iloc[i],
            data1["w0"].iloc[i],
        )

    # Calculate negative potential matrix
    neg_pot = np.zeros((N, m))
    sid = 0
    for be in leg2stateids:
        for te in range(mtempt):
            neg_pot[:, sid] = npot_fcn(
                e0=e0,
                epert=data1["pertE"],
                bet=bet[te],
                lam1=lambda1[be],
                lam2=lambda2[be],
                alpha=alpha[be],
                u0=u0[be],
                w0=w0[be],
            )
            sid += 1

    # Calculate state labels (reversed for leg2)
    statelabels = nstates - data1["stateid"]

    # Run UWHAM
    out = uwham_r(label=statelabels, logQ=neg_pot, ufactormax=1, ufactormin=1)

    # Reshape results
    ze = out["ze"].reshape(mtempt, mlam)
    ve = out["ve"].reshape(mtempt, mlam)

    # Calculate binding free energies
    dgbind2 = (-ze[:, -1] / bet) - (-ze[:, 0] / bet)
    ddgbind2 = np.sqrt(ve[:, -1] + ve[:, 0]) / bet

    dgb = dgbind1 - dgbind2
    ddgb = np.sqrt(ddgbind2 * ddgbind2 + ddgbind1 * ddgbind1)

    return dgb[0], ddgb[0], dgbind1[0], dgbind2[0], samplesperreplica


# if __name__ == "__main__":
#     print(
#         calculate_uwham(
#             rundir="tests/TYK2_A02_A09/TYK2_A02_A09_r0_1",
#             jobname="QB_A02_A09",
#             mintimeid=70,
#             maxtimeid=350,
#         )
#     )
#     print(
#         calculate_uwham(
#             rundir="tests/TYK2_A02_A09/TYK2_A02_A09_r0_2",
#             jobname="QB_A02_A09",
#             mintimeid=70,
#             maxtimeid=350,
#         )
#     )
#     print(
#         calculate_uwham(
#             rundir="tests/TYK2_A02_A09/TYK2_A02_A09_r0_3",
#             jobname="QB_A02_A09",
#             mintimeid=None,
#             maxtimeid=350,
#         )
#     )



import yaml
import json


def convert_cntl_to_yamljson(cntl_file, out_file):
    config = {}
    with open(cntl_file, "r") as f:
        for line in f:
            if line.startswith("#"):
                continue
            key, value = line.split("=")
            value = value.strip()
            if value.lower() in ("yes", "no"):
                value = {"yes": True, "no": False}[value.lower()]
            elif "," in value:
                if value.startswith('"'):
                    value = value[1:]
                if value.endswith('"'):
                    value = value[:-1]
                value = value.split(",")
                try:
                    value = [int(v.strip()) for v in value if len(v.strip()) > 0]
                except ValueError:
                    try:
                        value = [float(v.strip()) for v in value if len(v.strip()) > 0]
                    except ValueError:
                        pass
            else:
                try:
                    value = int(value.strip())
                except ValueError:
                    try:
                        value = float(value.strip())
                    except ValueError:
                        pass
            if key.strip().upper() in (
                "WALL_TIME",
                "CYCLE_TIME",
                "CHECKPOINT_TIME",
                "SUBJOBS_BUFFER_SIZE",
                "RE_SETUP",
                "JOB_TRANSPORT",
            ):
                continue
            config[key.strip()] = value

    with open(out_file, "w") as f:
        if out_file.endswith(".yaml"):
            yaml.dump(config, f, default_flow_style=None, sort_keys=False)
        elif out_file.endswith(".json"):
            json.dump(config, f)


if __name__ == "__main__":
    import sys

    convert_cntl_to_yamljson(sys.argv[1], sys.argv[2])



import os
import copy
import openmm.app as app
from openmm.unit import kelvin, kilojoules_per_mole, kilocalories_per_mole
from openmm import unit


class OMMReplica(object):
    #
    # Holds and manages OpenMM state for a replica
    #
    def __init__(self, replica_id, basename, worker, logger):
        self._id = replica_id
        self.basename = basename
        self.worker = worker
        self.context = worker.context
        self.ommsystem = worker.ommsystem
        self.logger = logger
        self.pot = None
        self.par = None
        self.cycle = 1
        self.stateid = None
        self.mdsteps = 0
        self.outfile = None

        state = self.context.getState(getPositions=True, getVelocities=True)
        self.positions = state.getPositions(asNumpy=True).value_in_unit(unit.nanometer)
        self.velocities = state.getVelocities(asNumpy=True).value_in_unit(
            unit.nanometer / unit.picosecond
        )

        if not os.path.isdir("r%d" % self._id):
            os.mkdir("r%d" % self._id)

        self.open_out()
        # may override stateid, positions, etc.
        self.load_checkpoint()
        self.open_dcd()

    def set_state(self, stateid, par):
        self.stateid = int(stateid)
        self.par = copy.deepcopy(par)

    def get_state(self):
        return (self.stateid, self.par)

    def get_energy(self):
        return self.pot

    def set_energy(self, pot):
        self.pot = copy.deepcopy(pot)

    def set_posvel(self, positions, velocities):
        self.positions[:] = positions
        self.velocities[:] = velocities

    def open_out(self):
        outfilename = "r%d/%s.out" % (self._id, self.basename)
        self.outfile = open(outfilename, "a+")
        if self.outfile is None:
            self.logger.warning("unable to open outfile %s" % outfilename)

    def load_checkpoint(self):
        ckptfile = "r%d/%s_ckpt.xml" % (self._id, self.basename)
        if os.path.isfile(ckptfile):
            self.logger.info("Loading checkpointfile %s" % ckptfile)
            self.worker.simulation.loadState(ckptfile)
            self.update_state_from_context()

    def save_checkpoint(self):
        ckptfile = "r%d/%s_ckpt.xml" % (self._id, self.basename)
        self.update_context_from_state()
        self.worker.simulation.saveState(ckptfile)

    def open_dcd(self):
        xtc = self.worker.config.get("XTC_TRAJECTORY", False)
        if xtc:
            filename = "r%d/%s.xtc" % (self._id, self.basename)
        else:
            filename = "r%d/%s.dcd" % (self._id, self.basename)
        append = os.path.isfile(filename)
        if append:
            mode = "r+b"
        else:
            mode = "wb"

        if xtc:
            self.traj = app.XTCFile(
                filename, self.worker.topology, self.ommsystem.MDstepsize, append=append
            )
        else:
            dcdfile = open(filename, mode)
            self.traj = app.DCDFile(
                dcdfile,
                self.worker.topology,
                self.ommsystem.MDstepsize,
                append=append,
            )
            dcdfile.flush()  # Force the writing of the DCD header

    def save_dcd(self):
        # TODO
        # boxsize options works only for NVT because the boxsize of the service worker
        # is not updated from the compute worker
        boxsize = self.worker.simulation.context.getState().getPeriodicBoxVectors()
        self.traj.writeModel(self.positions, periodicBoxVectors=boxsize)

    def set_mdsteps(self, mdsteps):
        self.mdsteps = mdsteps

    def get_mdsteps(self):
        return self.mdsteps

    def set_cycle(self, cycle):
        self.cycle = cycle

    def get_cycle(self):
        return self.cycle

    def get_stateid(self):
        return self.stateid


class OMMReplicaATM(OMMReplica):
    def save_out(self):
        # return the output data for current state as a tuple
        if self.pot is not None and self.par is not None:
            pot_energy = self.pot["potential_energy"]
            pert_energy = self.pot["perturbation_energy"]
            bias_energy = self.pot["bias_energy"]
            temperature = self.par["temperature"]
            lmbd1 = self.par["lambda1"]
            lmbd2 = self.par["lambda2"]
            alpha = self.par["alpha"]
            uh = self.par["uh"]
            w0 = self.par["w0"]
            direction = self.par["atmdirection"]

            replica_state_data = (
                self.stateid,
                temperature / kelvin,
                direction,
                lmbd1,
                lmbd2,
                alpha * kilocalories_per_mole,
                uh / kilocalories_per_mole,
                w0 / kilocalories_per_mole,
                pot_energy / kilocalories_per_mole,
                pert_energy / kilocalories_per_mole,
                bias_energy / kilocalories_per_mole,
            )

            return replica_state_data
        else:
            self.logger.error("unable to save output")
            raise ValueError

    def write_out(self, replica_state_data_list):
        # write list of saved output data to file
        if self.outfile is not None:
            for line in replica_state_data_list:
                self.outfile.write("%d %f %f %f %f %f %f %f %f %f %f\n" % line)

            self.outfile.flush()

        else:
            self.logger.error("unable to write output")
            raise IOError

    def update_state_from_context(self):
        self.cycle = int(self.context.getParameter(self.ommsystem.parameter["cycle"]))
        self.stateid = int(
            self.context.getParameter(self.ommsystem.parameter["stateid"])
        )
        self.mdsteps = int(
            self.context.getParameter(self.ommsystem.parameter["mdsteps"])
        )
        if self.par is None:
            self.par = {}
        self.par["temperature"] = (
            self.context.getParameter(self.ommsystem.parameter["temperature"]) * kelvin
        )
        self.par["lambda1"] = self.context.getParameter(
            self.ommsystem.atmforce.Lambda1()
        )
        self.par["lambda2"] = self.context.getParameter(
            self.ommsystem.atmforce.Lambda2()
        )
        self.par["alpha"] = (
            self.context.getParameter(self.ommsystem.atmforce.Alpha())
            / kilojoules_per_mole
        )
        self.par["uh"] = (
            self.context.getParameter(self.ommsystem.atmforce.Uh())
            * kilojoules_per_mole
        )
        self.par["w0"] = (
            self.context.getParameter(self.ommsystem.atmforce.W0())
            * kilojoules_per_mole
        )
        self.par["atmdirection"] = self.context.getParameter(
            self.ommsystem.atmforce.Direction()
        )
        self.par["atmintermediate"] = self.context.getParameter(
            self.ommsystem.parameter["atmintermediate"]
        )
        self.par[self.ommsystem.atmforce.Umax()] = (
            self.context.getParameter(self.ommsystem.atmforce.Umax())
            * kilojoules_per_mole
        )
        self.par[self.ommsystem.atmforce.Ubcore()] = (
            self.context.getParameter(self.ommsystem.atmforce.Ubcore())
            * kilojoules_per_mole
        )
        self.par[self.ommsystem.atmforce.Acore()] = self.context.getParameter(
            self.ommsystem.atmforce.Acore()
        )
        if self.pot is None:
            self.pot = {}
        self.pot["potential_energy"] = (
            self.context.getParameter(self.ommsystem.parameter["potential_energy"])
            * kilojoules_per_mole
        )
        self.pot["perturbation_energy"] = (
            self.context.getParameter(self.ommsystem.parameter["perturbation_energy"])
            * kilojoules_per_mole
        )
        self.pot["bias_energy"] = (
            self.context.getParameter(self.ommsystem.parameter["bias_energy"])
            * kilojoules_per_mole
        )
        state = self.context.getState(getPositions=True, getVelocities=True)
        self.positions = state.getPositions(asNumpy=True)
        self.velocities = state.getVelocities(asNumpy=True)

    def update_context_from_state(self):
        self.context.setParameter(self.ommsystem.parameter["cycle"], self.cycle)
        self.context.setParameter(self.ommsystem.parameter["stateid"], self.stateid)
        self.context.setParameter(self.ommsystem.parameter["mdsteps"], self.mdsteps)
        if self.par is not None:
            self.context.setParameter(
                self.ommsystem.parameter["temperature"],
                self.par["temperature"] / kelvin,
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Lambda1(), self.par["lambda1"]
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Lambda2(), self.par["lambda2"]
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Alpha(), self.par["alpha"] * kilojoules_per_mole
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Uh(), self.par["uh"] / kilojoules_per_mole
            )
            self.context.setParameter(
                self.ommsystem.atmforce.W0(), self.par["w0"] / kilojoules_per_mole
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Direction(), self.par["atmdirection"]
            )
            self.context.setParameter(
                self.ommsystem.parameter["atmintermediate"], self.par["atmintermediate"]
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Umax(),
                self.par[self.ommsystem.atmforce.Umax()] / kilojoules_per_mole,
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Ubcore(),
                self.par[self.ommsystem.atmforce.Ubcore()] / kilojoules_per_mole,
            )
            self.context.setParameter(
                self.ommsystem.atmforce.Acore(),
                self.par[self.ommsystem.atmforce.Acore()],
            )
        if self.pot is not None:
            self.context.setParameter(
                self.ommsystem.parameter["potential_energy"],
                self.pot["potential_energy"] / kilojoules_per_mole,
            )
            self.context.setParameter(
                self.ommsystem.parameter["perturbation_energy"],
                self.pot["perturbation_energy"] / kilojoules_per_mole,
            )
            self.context.setParameter(
                self.ommsystem.parameter["bias_energy"],
                self.pot["bias_energy"] / kilojoules_per_mole,
            )
        self.context.setPositions(self.positions)
        self.context.setVelocities(self.velocities)



import shutil
import os


curr_dir = os.path.dirname(os.path.abspath(__file__))


def _read_dcd_header(file):
    import struct

    with open(file, "r+b") as f:
        f.seek(8, os.SEEK_SET)
        numFrames = struct.unpack("<i", f.read(4))[0]
        f.seek(20, os.SEEK_SET)
        currStep = struct.unpack("<i", f.read(4))[0]
        return numFrames, currStep


def _test_structprep(tmp_path):
    from atm.rbfe_structprep import rbfe_structprep

    shutil.copytree(
        os.path.join(curr_dir, "QB_A08_A07"), os.path.join(tmp_path, "QB_A08_A07")
    )
    rbfe_structprep(os.path.join(tmp_path, "QB_A08_A07", "QB_A08_A07_asyncre.yaml"))


def _test_production(tmp_path):
    from atm.rbfe_production import rbfe_production

    shutil.copytree(
        os.path.join(curr_dir, "QB_A08_A07_completed"),
        os.path.join(tmp_path, "QB_A08_A07_completed"),
    )
    rbfe_production(
        os.path.join(tmp_path, "QB_A08_A07_completed", "QB_A08_A07_asyncre.yaml")
    )
    for i in range(4):
        assert os.path.exists(
            os.path.join(tmp_path, "QB_A08_A07_completed", f"r{i}", "QB_A08_A07.dcd")
        )


def _test_production_xtc(tmp_path):
    from atm.rbfe_production import rbfe_production
    import yaml

    shutil.copytree(
        os.path.join(curr_dir, "QB_A08_A07_completed"),
        os.path.join(tmp_path, "QB_A08_A07_completed"),
    )

    configfile = os.path.join(
        tmp_path, "QB_A08_A07_completed", "QB_A08_A07_asyncre.yaml"
    )
    with open(configfile, "r") as f:
        config = yaml.safe_load(f)
    config["XTC_TRAJECTORY"] = True
    with open(configfile, "w") as f:
        yaml.dump(config, f)

    rbfe_production(configfile)
    for i in range(4):
        assert os.path.exists(
            os.path.join(tmp_path, "QB_A08_A07_completed", f"r{i}", "QB_A08_A07.xtc")
        )


def _test_production_incremental(tmp_path):
    from atm.rbfe_production import rbfe_production
    import yaml

    shutil.copytree(
        os.path.join(curr_dir, "QB_A08_A07_completed"),
        os.path.join(tmp_path, "QB_A08_A07_completed"),
    )

    configfile = os.path.join(
        tmp_path, "QB_A08_A07_completed", "QB_A08_A07_asyncre.yaml"
    )
    with open(configfile, "r") as f:
        config = yaml.safe_load(f)
    config["MAX_SAMPLES"] = "+1"
    with open(configfile, "w") as f:
        yaml.dump(config, f)

    rbfe_production(configfile)
    for i in range(4):
        dcdf = os.path.join(tmp_path, "QB_A08_A07_completed", f"r{i}", "QB_A08_A07.dcd")
        numFrames, _ = _read_dcd_header(dcdf)
        assert numFrames == 1

    startsampl_file = os.path.join(tmp_path, "QB_A08_A07_completed", "starting_sample")
    with open(startsampl_file, "r") as f:
        starting_sample = int(f.read().strip())
        assert starting_sample == 1
    prog_file = os.path.join(tmp_path, "QB_A08_A07_completed", "progress")
    with open(prog_file, "r") as f:
        progress = float(f.read().strip())
        assert progress == 0.0

    # Run another 2 samples
    os.remove(startsampl_file)
    os.remove(prog_file)
    config["MAX_SAMPLES"] = "+2"
    with open(configfile, "w") as f:
        yaml.dump(config, f)
    rbfe_production(configfile)
    for i in range(4):
        dcdf = os.path.join(tmp_path, "QB_A08_A07_completed", f"r{i}", "QB_A08_A07.dcd")
        numFrames, _ = _read_dcd_header(dcdf)
        assert numFrames == 2

    with open(startsampl_file, "r") as f:
        starting_sample = int(f.read().strip())
        assert starting_sample == 2
    with open(prog_file, "r") as f:
        progress = float(f.read().strip())
        assert progress == 0.5


def _test_uwham_analysis(tmp_path):
    from atm.uwham import calculate_uwham

    shutil.copytree(
        os.path.join(curr_dir, "TYK2_A02_A09"), os.path.join(tmp_path, "TYK2_A02_A09")
    )

    run_dir = os.path.join(tmp_path, "TYK2_A02_A09", "TYK2_A02_A09_r0_1")
    ddG, ddG_std, dgbind1, dgbind2, samples = calculate_uwham(run_dir, "QB_A02_A09", 70)

    expected_ddG = -1.0582613105156682
    expected_ddG_std = 0.1967262793670161
    assert abs(ddG - expected_ddG) < 1e-8
    assert abs(ddG_std - expected_ddG_std) < 1e-8
    assert samples == 351
    assert abs(dgbind1 - 16.12549393024899) < 1e-8
    assert abs(dgbind2 - 17.183755240764658) < 1e-8

    run_dir = os.path.join(tmp_path, "TYK2_A02_A09", "TYK2_A02_A09_r0_2")
    ddG, ddG_std, dgbind1, dgbind2, samples = calculate_uwham(run_dir, "QB_A02_A09", 70)

    expected_ddG = -1.9492444343451716
    expected_ddG_std = 0.18548486855089807
    assert abs(ddG - expected_ddG) < 1e-8
    assert abs(ddG_std - expected_ddG_std) < 1e-8
    assert samples == 421
    assert abs(dgbind1 - 15.690178371190054) < 1e-8
    assert abs(dgbind2 - 17.639422805535226) < 1e-8

    run_dir = os.path.join(tmp_path, "TYK2_A02_A09", "TYK2_A02_A09_r0_3")
    ddG, ddG_std, dgbind1, dgbind2, samples = calculate_uwham(
        run_dir, "QB_A02_A09", None
    )

    expected_ddG = -1.2625416487680994
    expected_ddG_std = 0.2068400471106031
    assert abs(ddG - expected_ddG) < 1e-8
    assert abs(ddG_std - expected_ddG_std) < 1e-8
    assert samples == 480
    assert abs(dgbind1 - 16.159606073746524) < 1e-8
    assert abs(dgbind2 - 17.422147722514623) < 1e-8



from openmm import NonbondedForce
from openmm.unit import elementary_charge
from openmmml.mlpotential import MLPotential, MLPotentialImpl, MLPotentialImplFactory
import openmmtorch
import torch as pt
from torchmdnet.models.model import load_model


class TorchMDNETForce(pt.nn.Module):

    def __init__(
        self,
        model_file,
        atomic_numbers,
        group_indices,
        total_charges,
        max_num_neighbors,
    ):
        super().__init__()
        assert len(group_indices) == len(total_charges)

        self.model = load_model(
            model_file,
            derivative=False,
            max_num_neighbors=max_num_neighbors,
            remove_ref_energy=False,
            static_shapes=True,
            check_errors=False,
        )
        for parameter in self.model.parameters():
            parameter.requires_grad = False

        self.register_buffer(
            "all_atom_indices", pt.tensor(sum(group_indices, []), dtype=pt.long)
        )
        self.register_buffer(
            "atomic_numbers",
            pt.tensor(atomic_numbers, dtype=pt.long)[self.all_atom_indices],
        )
        batch = sum(
            [[i] * len(atom_indices) for i, atom_indices in enumerate(group_indices)],
            [],
        )
        self.register_buffer("batch", pt.tensor(batch, dtype=pt.long))
        self.register_buffer("total_charges", pt.tensor(total_charges, dtype=pt.long))

    def forward(self, positions):
        positions = (
            pt.index_select(positions, 0, self.all_atom_indices).to(pt.float32) * 10
        )  # nm --> A
        energies, *_ = self.model(
            self.atomic_numbers, positions, batch=self.batch, q=self.total_charges
        )
        return energies.sum() * 96.4915666370759  # eV -> kJ/mol


class TorchMDNETImpl(MLPotentialImpl):

    def __init__(
        self, name, model_file, group_indices, max_num_neighbors, use_cuda_graphs
    ):
        self.name = name
        self.model_file = model_file
        self.group_indices = group_indices
        self.max_num_neighbors = int(max_num_neighbors)
        self.use_cuda_graphs = bool(use_cuda_graphs)

    @staticmethod
    def _get_total_charge(system, atom_indices):
        for force in system.getForces():
            if isinstance(force, NonbondedForce):
                return round(
                    sum(
                        force.getParticleParameters(i)[0].value_in_unit(
                            elementary_charge
                        )
                        for i in atom_indices
                    )
                )

    def addForces(self, topology, system, all_atom_indices, force_group):
        atomic_numbers = [atom.element.atomic_number for atom in topology.atoms()]
        total_charges = [
            self._get_total_charge(system, atom_indices)
            for atom_indices in self.group_indices
        ]
        assert sum(self.group_indices, []) == all_atom_indices

        force = TorchMDNETForce(
            self.model_file,
            atomic_numbers,
            self.group_indices,
            total_charges,
            self.max_num_neighbors,
        )
        force = openmmtorch.TorchForce(pt.jit.script(force))
        force.setProperty("useCUDAGraphs", "true" if self.use_cuda_graphs else "false")
        force.setForceGroup(force_group)
        system.addForce(force)


class TorchMDNETImplFactory(MLPotentialImplFactory):

    def createImpl(
        self, name, model_file, group_indices, max_num_neighbors, use_cuda_graphs=True
    ):
        return TorchMDNETImpl(
            name, model_file, group_indices, max_num_neighbors, use_cuda_graphs
        )


MLPotential.registerImplFactory("TorchMD-NET", TorchMDNETImplFactory())
